import numpy as np
from matplotlib import pyplot as plt
from matplotlib.dates import HourLocator, DateFormatter

from skyfield.api import load, EarthSatellite, wgs84

from skyfield.api import load, wgs84

starlink_url = "https://celestrak.com/NORAD/elements/supplemental/starlink.txt"
starlinks = load.tle_file(starlink_url)

# Build the time range `t` over which to plot, plus other values.

ts = load.timescale()

# Load the satellite's TLE entry.

astra2e = EarthSatellite(
    "1 39285U 13056A   22137.15428389  .00000157  00000+0  00000+0 0  9998",
    "2 39285   0.0686 331.1550 0002286  66.0071 281.6942  1.00269260 31625",
    "ASTRA 2E",
    ts,
)
astra2f = EarthSatellite(
    "1 38778U 12051A   22142.82294109  .00000140  00000+0  00000+0 0  9991",
    "2 38778   0.0565 240.3451 0000453 298.7811  25.7511  1.00269234 26228",
    "ASTRA 2F",
    ts,
)

astra2g = EarthSatellite(
    "1 40364U 14089A   22142.53592181  .00000140  00000+0  00000+0 0  9997",
    "2 40364   0.0579  63.3703 0005136 355.6036  42.3018  1.00273360 27060",
    "ASTRA 2G",
    ts,
)

ISS = EarthSatellite(
    "1 25544U 98067A   22143.37738542  .00010371  00000+0  19191-3 0  9996",
    "2 25544  51.6438  99.4639 0004967 147.2337 318.3670 15.49625671341330",
    "ISS (ZARYA)",
    ts,
)
satellites = [astra2e, astra2f, astra2g, ISS]
numberofstarlinks = len(starlinks) # total of 2386
for sat in range(numberofstarlinks):
    satellites.append(starlinks[sat])
# Compute geocentric positions for the satellite.
RoyalObservatoryGreenwich = wgs84.latlon(+51.476852, -0.0005, 45)

t0 = ts.utc(2022, 1, 23)
t1 = ts.utc(2022, 1, 24)

time_points = 5 * 60   # track over 5 hours
# change over time
t = ts.utc(2022, 5, 23, 13, range(0, time_points))
timebase = t.utc_datetime()

# satellites=[satellite]
# do for all starlink satellites
tracks = []
for sat in satellites:
    pos = sat.at(t)
    tracks.append(pos)
    latcheck = np.any(np.isnan(wgs84.subpoint_of(pos).latitude.degrees))
    loncheck = np.any(np.isnan(wgs84.subpoint_of(pos).longitude.degrees))
    if latcheck | loncheck:
        tracks.pop(-1)
        print("caught one")

# map reference site to map satellite paths relative to the reference site
import folium

m = folium.Map(
    location=[
        RoyalObservatoryGreenwich.latitude.degrees,
        RoyalObservatoryGreenwich.longitude.degrees,
    ],
    zoom_start=2,
    tiles="Stamen Terrain",
)

folium.Marker(
    [
        RoyalObservatoryGreenwich.latitude.degrees,
        RoyalObservatoryGreenwich.longitude.degrees,
    ],
    tooltip="Royal Observatory Greenwich",
    icon=folium.Icon(
        icon="fa-circle", prefix="fa", color="red"  # Marker for Measurement Site
    ),
).add_to(m)

# %%
# GeoJSON
# ~~~~~~~
# The Folium plugin `TimestampedGeoJson`_ will be used to plot our tracks using timestamped
# GeoJSONs. As a result, we want to convert our data into `GeoJSON format`_. Firstly, we create
# our feature collection which we will append our features to.
#
# .. _GeoJSON format: https://geojson.org/
# .. _TimestampedGeoJson:  https://python-visualization.github.io/folium/plugins.html

geo_features = list()
geo_json = {
    "type": "FeatureCollection",
    "features": geo_features,
}
# %%
# Plotting Tracks
# ~~~~~~~~~~~~~~~
# Now we append our tracks to our feature collection list. We define `colour_iter` which will
# allow us to cycle through different colours when mapping our tracks.

from collections import defaultdict
from itertools import cycle

colour_iter = iter(
    cycle(
        [
            "red",
            "blue",
            "green",
            "purple",
            "orange",
            "darkred",
            "#0909FF",
            "#F70D1A",
            "#FF6700",
            "lightgreen",
            "#0AFFFF",
            "#12AD2B",
            "#E2F516",
            "#FFFF00",
            "#F52887",
        ]
    )
)
colour = defaultdict(lambda: next(colour_iter))

trail_Size = 7  # trail_Size is the number of timestamps we want track to trail for
from io import BytesIO
import base64

png = "./satelliteicon.png"
with open(png, "rb") as lf:
    # open in binary mode, read bytes, encode, decode obtained bytes as utf-8 string
    b64_content = base64.b64encode(lf.read()).decode("utf-8")
# we only care for sattelites above the horizon from the point of interest, in this example the Royal Observatory, so eliminate plot points with altitude of less than 0 degrees.
# I have no idea why, but for folium line points and markers, the normal order of latitude,longitude is reversed, unless I am missing a config change.
for track in range(len(tracks)):
    lat = wgs84.subpoint_of(tracks[track]).latitude
    lon = wgs84.subpoint_of(tracks[track]).longitude
    relative_measure = (satellites[track] - RoyalObservatoryGreenwich).at(t)
    alt, az, distance = relative_measure.altaz()
    altitude = alt.degrees
    azimuth = az.degrees
    relative_range = distance.km
    plot_points = []
    plot_times = []
    az_trim = []
    alt_trim = []
    range_trim = []
    for point in range(len(lat.degrees)):
        if altitude[point] >= 0.0:
            plot_points.append((lon.degrees[point], lat.degrees[point]))
            plot_times.append(timebase[point].strftime("%Y-%m-%d %H:%M:%S"))
            az_trim.append(azimuth[point])
            alt_trim.append(altitude[point])
            range_trim.append(relative_range[point])

    for time_index, time in enumerate(plot_times):
        geo_features.append(
            {
                "type": "Feature",
                "properties": {
                    "name": satellites[track].name,
                    "style": {"color": colour[track], "weight": 6},
                    "times": [time] * len(plot_points[: time_index + 1][-trail_Size:]),
                },
                "geometry": {
                    "type": "LineString",
                    "coordinates": plot_points[: time_index + 1][-trail_Size:],
                },
            }
        )
        geo_features.append(
            {
                "type": "Feature",
                "properties": {
                    "icon": "marker",
                    "iconstyle": {
                        "iconUrl": f"http://icons.iconarchive.com/icons/google/noto-emoji-travel-places/1024/42597-satellite-icon.png",
                        "iconSize": [24, 24],
                        "fillOpacity": 1,
                        "popupAnchor": [1, -17],
                    },
                    "popup": "Satellite: " + satellites[track].name + "<dd>"
                    "Latitude: "
                    + "%s" % float("%.8g" % plot_points[time_index][0])
                    + "<dd>"
                    "Longitude: "
                    + "%s" % float("%.8g" % plot_points[time_index][1])
                    + "<dd>"  # rounding 8 sigfigs
                    "Azimuth: "
                    + "%s" % float("%.8g" % az_trim[time_index])
                    + "°"
                    + "<dd>"
                    "Altitude: "
                    + "%s" % float("%.8g" % alt_trim[time_index])
                    + "°"
                    + "<dd>"
                    "Slant Range: "
                    + "%s" % float("%.8g" % range_trim[time_index])
                    + "km",
                    "name": satellites[track].name,
                    "style": {"color": "black", "weight": 2},
                    "times": [time],
                },
                "geometry": {
                    "type": "MultiPoint",
                    "coordinates": [plot_points[time_index]],
                },
            }
        )

# %%
# The Results
# ~~~~~~~~~~~

from folium.plugins import TimestampedGeoJson, Fullscreen

Fullscreen().add_to(m)

TimestampedGeoJson(
    data=geo_json,
    transition_time=200,
    auto_play=True,
    add_last_point=False,
    period="PT1M",
    duration="PT1S",
).add_to(m)

# %%

m.save("satellitetracks.html")



import numpy as np
import open3d as o3d
import scipy.constants
import copy

from ..base_types import scattering_t
from ..electromagnetics import empropagation as EM
from ..geometry import targets as TL
from ..raycasting import rayfunctions as RF
from ..geometry import geometryfunctions as GF

def calculate_scattering(
    aperture_coords,
    sink_coords,
    excitation_function,
    antenna_solid,
    desired_E_axis,
    scatter_points=None,
    wavelength=1.0,
    scattering=0,
    elements=False,
    sampling_freq=1e9,
    los=False,
    num_samples=10000,
    mesh_resolution=0.5,
    antenna_axes=np.eye(3),
    project_vectors=True
):
    """
    Based upon the parameters given, calculate the time domain scattering for the apertures and sinks.
    The defaults for sampling time are based upon a sampling rate of 1GHz, and a sample time of 1ms.

    Parameters
    ----------
    aperture_coords : :class:`open3d.geometry.TriangleMesh`
        source coordinates
    sink_coords : :class:`open3d.geometry.TriangleMesh`
        sink coordinates
    antenna_solid : :class:`lyceanem.base_classes.structures`
        the class should contain all the environment for scattering, providing the blocking for the rays
    desired_E_axis : 1D numpy array of floats
        the desired excitation vector, can be a 1*3 array or a n*3 array if multiple different exciations are desired in one lauch
    scatter_points : :class:`open3d.geometry.TriangleMesh`
        the scattering points in the environment. Defaults to [None], in which case scattering points will be generated from the antenna_solid. If no scattering should be considered then set scattering to [0].
    wavelength : float
        the wavelength of interest in metres
    scattering : int
        the number of reflections to be considered, defaults to [0], but up to 2 can be considered. The higher this number to greater to computational effort, and for most situations 1 should be ample.
    elements : boolean
        whether the sources and sinks should be considered as elements of a phased array, or a fixed phase aperture like a horn or reflector
    sampling_freq : float
        the desired sampling frequency, should be twice the highest frequency of interest to avoid undersampling artifacts
    num_samples : int
        the length of the desired sampling, can be calculated from the desired model time
    mesh_resolution : float
        the desired mesh resolution in terms of wavelengths if scattering points are not provided. A scattering mesh is generated on the surfaces of all provided trianglemesh structures.

    Return
    -------
    Ex : numpy array of float
        the x directed voltage at the sink coordinates in the time domain, if elements=True, then it will be an array of num_sinks * num_samples, otherwise it will be a 1D array
    Ey : numpy array of float
        the y directed voltage at the sink coordinates in the time domain, if elements=True, then it will be an array of num_sinks * num_samples, otherwise it will be a 1D array
    Ez : numpy array of float
        the z directed voltage at the sink coordinates in the time domain, if elements=True, then it will be an array of num_sinks * num_samples, otherwise it will be a 1D array
    Waketimes : numpy array of float
        the shortest time required for a ray to reach any sink from any source, as long as elements=False. If elements=True, then this is not implemented in the same way, and will return the shortest time required for the final source.
    """

    time_index = np.linspace(0, num_samples / sampling_freq, num_samples)
    multiE = False

    num_sources = len(np.asarray(aperture_coords.points))
    num_sinks = len(np.asarray(sink_coords.points))
    environment_triangles=GF.mesh_conversion(antenna_solid)

    if not multiE:
        if project_vectors:
            conformal_E_vectors = EM.calculate_conformalVectors(
                desired_E_axis, np.asarray(aperture_coords.normals), antenna_axes
            )
        else:
            if desired_E_axis.shape[0] == np.asarray(aperture_coords.normals).shape[0]:
                conformal_E_vectors = copy.deepcopy(desired_E_axis)
            else:
                conformal_E_vectors = np.repeat(
                    desired_E_axis.reshape(1, 3).astype(np.complex64), num_sources, axis=0
                )
    else:
        if project_vectors:
            conformal_E_vectors = EM.calculate_conformalVectors(
                desired_E_axis, np.asarray(aperture_coords.normals), antenna_axes
            )
        else:
            if desired_E_axis.size == 3:
                conformal_E_vectors = np.repeat(
                    desired_E_axis[0, :].astype(np.float32), num_sources, axis=0
                ).reshape(num_sources, 3)
            else:
                conformal_E_vectors = desired_E_axis.reshape(num_sources, 3)

    if scattering == 0:
        # only use the aperture point cloud, no scattering required.
        scatter_points = o3d.geometry.PointCloud()

        unified_model = np.append(
            np.asarray(aperture_coords.points).astype(np.float32),
            np.asarray(sink_coords.points).astype(np.float32),
            axis=0,
        )
        unified_normals = np.append(
            np.asarray(aperture_coords.normals).astype(np.float32),
            np.asarray(sink_coords.normals).astype(np.float32),
            axis=0,
        )
        unified_weights = np.ones((unified_model.shape[0], 3), dtype=np.complex64)
        unified_weights[0:num_sources, :] = (
            conformal_E_vectors #/ num_sources
        )  # set total amplitude to 1 for the aperture
        unified_weights[num_sources : num_sources + num_sinks, :] = (
            1 #/ num_sinks
        )  # set total amplitude to 1 for the aperture
        point_informationv2 = np.empty((len(unified_model)), dtype=scattering_t)
        # set all sources as magnetic current sources, and permittivity and permeability as free space
        point_informationv2[:]["Electric"] = True
        point_informationv2[:]["permittivity"] = 8.8541878176e-12
        point_informationv2[:]["permeability"] = 1.25663706212e-6
        # set position, velocity, normal, and weight of sources
        point_informationv2[0:num_sources]["px"] = np.asarray(
            aperture_coords.points
        ).astype(np.float32)[:, 0]
        point_informationv2[0:num_sources]["py"] = np.asarray(
            aperture_coords.points
        ).astype(np.float32)[:, 1]
        point_informationv2[0:num_sources]["pz"] = np.asarray(
            aperture_coords.points
        ).astype(np.float32)[:, 2]
        point_informationv2[0:num_sources]["nx"] = np.asarray(
            aperture_coords.normals
        ).astype(np.float32)[:, 0]
        point_informationv2[0:num_sources]["ny"] = np.asarray(
            aperture_coords.normals
        ).astype(np.float32)[:, 1]
        point_informationv2[0:num_sources]["nz"] = np.asarray(
            aperture_coords.normals
        ).astype(np.float32)[:, 2]
        # set position and velocity of sinks
        point_informationv2[num_sources : (num_sources + num_sinks)]["px"] = np.asarray(
            sink_coords.points
        ).astype(np.float32)[:, 0]
        point_informationv2[num_sources : (num_sources + num_sinks)]["py"] = np.asarray(
            sink_coords.points
        ).astype(np.float32)[:, 1]
        point_informationv2[num_sources : (num_sources + num_sinks)]["pz"] = np.asarray(
            sink_coords.points
        ).astype(np.float32)[:, 2]
        point_informationv2[num_sources : (num_sources + num_sinks)]["nx"] = np.asarray(
            sink_coords.normals
        ).astype(np.float32)[:, 0]
        point_informationv2[num_sources : (num_sources + num_sinks)]["ny"] = np.asarray(
            sink_coords.normals
        ).astype(np.float32)[:, 1]
        point_informationv2[num_sources : (num_sources + num_sinks)]["nz"] = np.asarray(
            sink_coords.normals
        ).astype(np.float32)[:, 2]

        point_informationv2[:]["ex"] = unified_weights[:, 0]
        point_informationv2[:]["ey"] = unified_weights[:, 1]
        point_informationv2[:]["ez"] = unified_weights[:, 2]
        scatter_mask = np.zeros((point_informationv2.shape[0]), dtype=np.int32)
        scatter_mask[0:num_sources] = 0
        scatter_mask[(num_sources + num_sinks) :] = 0

    else:
        # create scatter points on antenna solids based upon a half wavelength square
        if scatter_points is None:
            scatter_points, areas = TL.source_cloud_from_shape(
                antenna_solid, 1e-6, (wavelength * mesh_resolution) ** 2
            )

        unified_model = np.append(
            np.append(
                np.asarray(aperture_coords.points).astype(np.float32),
                np.asarray(sink_coords.points).astype(np.float32),
                axis=0,
            ),
            np.asarray(scatter_points.points).astype(np.float32),
            axis=0,
        )
        unified_normals = np.append(
            np.append(
                np.asarray(aperture_coords.normals).astype(np.float32),
                np.asarray(sink_coords.normals).astype(np.float32),
                axis=0,
            ),
            np.asarray(scatter_points.normals).astype(np.float32),
            axis=0,
        )
        unified_weights = np.ones((unified_model.shape[0], 3), dtype=np.complex64)
        unified_weights[0:num_sources, :] = (
            conformal_E_vectors #/ num_sources
        )  # set total amplitude to 1 for the aperture
        unified_weights[num_sources : num_sources + num_sinks, :] = (
            1 #/ num_sinks
        )  # set total amplitude to 1 for the aperture
        unified_weights[num_sources + num_sinks :, :] = 1 #/ len(
          #  np.asarray(scatter_points.points)
        #)  # set total amplitude to 1 for the aperture
        point_informationv2 = np.empty((len(unified_model)), dtype=scattering_t)
        # set all sources as magnetic current sources, and permittivity and permeability as free space
        point_informationv2[:]["Electric"] = True
        point_informationv2[:]["permittivity"] = 8.8541878176e-12
        point_informationv2[:]["permeability"] = 1.25663706212e-6
        # set position, velocity, normal, and weight of sources
        point_informationv2[0:num_sources]["px"] = np.asarray(
            aperture_coords.points
        ).astype(np.float32)[:, 0]
        point_informationv2[0:num_sources]["py"] = np.asarray(
            aperture_coords.points
        ).astype(np.float32)[:, 1]
        point_informationv2[0:num_sources]["pz"] = np.asarray(
            aperture_coords.points
        ).astype(np.float32)[:, 2]
        point_informationv2[0:num_sources]["nx"] = np.asarray(
            aperture_coords.normals
        ).astype(np.float32)[:, 0]
        point_informationv2[0:num_sources]["ny"] = np.asarray(
            aperture_coords.normals
        ).astype(np.float32)[:, 1]
        point_informationv2[0:num_sources]["nz"] = np.asarray(
            aperture_coords.normals
        ).astype(np.float32)[:, 2]
        # point_informationv2[0:num_sources]['ex']=unified_weights[0:num_sources,0]
        # point_informationv2[0:num_sources]['ey']=unified_weights[0:num_sources,1]
        # point_informationv2[0:num_sources]['ez']=unified_weights[0:num_sources,2]
        # set position and velocity of sinks
        point_informationv2[num_sources : (num_sources + num_sinks)]["px"] = np.asarray(
            sink_coords.points
        ).astype(np.float32)[:, 0]
        point_informationv2[num_sources : (num_sources + num_sinks)]["py"] = np.asarray(
            sink_coords.points
        ).astype(np.float32)[:, 1]
        point_informationv2[num_sources : (num_sources + num_sinks)]["pz"] = np.asarray(
            sink_coords.points
        ).astype(np.float32)[:, 2]
        # point_informationv2[num_sources:(num_sources+num_sinks)]['vx']=0.0
        # point_informationv2[num_sources:(num_sources+num_sinks)]['vy']=0.0
        # point_informationv2[num_sources:(num_sources+num_sinks)]['vz']=0.0
        point_informationv2[num_sources : (num_sources + num_sinks)]["nx"] = np.asarray(
            sink_coords.normals
        ).astype(np.float32)[:, 0]
        point_informationv2[num_sources : (num_sources + num_sinks)]["ny"] = np.asarray(
            sink_coords.normals
        ).astype(np.float32)[:, 1]
        point_informationv2[num_sources : (num_sources + num_sinks)]["nz"] = np.asarray(
            sink_coords.normals
        ).astype(np.float32)[:, 2]
        point_informationv2[(num_sources + num_sinks) :]["px"] = np.asarray(
            scatter_points.points
        ).astype(np.float32)[:, 0]
        point_informationv2[(num_sources + num_sinks) :]["py"] = np.asarray(
            scatter_points.points
        ).astype(np.float32)[:, 1]
        point_informationv2[(num_sources + num_sinks) :]["pz"] = np.asarray(
            scatter_points.points
        ).astype(np.float32)[:, 2]
        point_informationv2[(num_sources + num_sinks) :]["nx"] = np.asarray(
            scatter_points.normals
        ).astype(np.float32)[:, 0]
        point_informationv2[(num_sources + num_sinks) :]["ny"] = np.asarray(
            scatter_points.normals
        ).astype(np.float32)[:, 1]
        point_informationv2[(num_sources + num_sinks) :]["nz"] = np.asarray(
            scatter_points.normals
        ).astype(np.float32)[:, 2]
        point_informationv2[:]["ex"] = unified_weights[:, 0]
        point_informationv2[:]["ey"] = unified_weights[:, 1]
        point_informationv2[:]["ez"] = unified_weights[:, 2]
        # scatter_mask = np.zeros((point_informationv2.shape[0]), dtype=np.int32)
        # scatter_mask[0:num_sources] = 0
        # scatter_mask[(num_sources + num_sinks):] = scattering

    # full_index, initial_index = RF.integratedraycastersetup(num_sources,
    #                                                        num_sinks,
    #                                                        point_informationv2,
    #                                                        RF.convertTriangles(antenna_solid),
    #                                                        scatter_mask)
    full_index, rays = RF.workchunkingv2(
        np.asarray(aperture_coords.points).astype(np.float32),
        np.asarray(sink_coords.points).astype(np.float32),
        np.asarray(scatter_points.points).astype(np.float32),
        environment_triangles,
        scattering + 1,
        line_of_sight=los,
    )

    if not elements:
        # create efiles for model
        if multiE:
            Ex = np.zeros((desired_E_axis.shape[1], num_samples), dtype=np.float64)
            Ey = np.zeros((desired_E_axis.shape[1], num_samples), dtype=np.float64)
            Ez = np.zeros((desired_E_axis.shape[1], num_samples), dtype=np.float64)
            for e_inc in range(desired_E_axis.shape[0]):
                conformal_E_vectors = EM.calculate_conformalVectors(
                    desired_E_axis[e_inc, :],
                    np.asarray(aperture_coords.normals).astype(np.float32),
                )
                unified_weights[0:num_sources, :] = conformal_E_vectors #/ num_sources
                point_informationv2[:]["ex"] = unified_weights[:, 0]
                point_informationv2[:]["ey"] = unified_weights[:, 1]
                point_informationv2[:]["ez"] = unified_weights[:, 2]
                scattering_coefficient = 1 / 4 * scipy.constants.pi
                TimeMap, WakeTimes = EM.TimeDomainv3(
                    num_sources,
                    num_sinks,
                    point_informationv2,
                    full_index,
                    scattering_coefficient,
                    wavelength,
                    excitation_function,
                    sampling_freq,
                    num_samples,
                )
                wake_index = np.digitize(WakeTimes, time_index)
                Ex[e_inc, wake_index:] = np.dot(
                    np.ones((num_sources)),
                    np.dot(
                        np.ones((num_sinks)),
                        TimeMap[:, :, : (num_samples - wake_index), 0],
                    ),
                )
                Ey[e_inc, wake_index:] = np.dot(
                    np.ones((num_sources)),
                    np.dot(
                        np.ones((num_sinks)),
                        TimeMap[:, :, : (num_samples - wake_index), 1],
                    ),
                )
                Ez[e_inc, wake_index:] = np.dot(
                    np.ones((num_sources)),
                    np.dot(
                        np.ones((num_sinks)),
                        TimeMap[:, :, : (num_samples - wake_index), 2],
                    ),
                )

        else:
            Ex = np.zeros((num_samples), dtype=np.float64)
            Ey = np.zeros((num_samples), dtype=np.float64)
            Ez = np.zeros((num_samples), dtype=np.float64)
            scattering_coefficient = 1 / 4 * scipy.constants.pi
            TimeMap, WakeTimes = EM.TimeDomainv3(
                num_sources,
                num_sinks,
                point_informationv2,
                full_index,
                scattering_coefficient,
                wavelength,
                excitation_function,
                sampling_freq,
                num_samples,
            )
            wake_index = np.digitize(WakeTimes, time_index)
            Ex[wake_index:] = np.dot(
                np.ones((num_sources)),
                np.dot(
                    np.ones((num_sinks)), TimeMap[:, :, : (num_samples - wake_index), 0]
                ),
            )
            Ey[wake_index:] = np.dot(
                np.ones((num_sources)),
                np.dot(
                    np.ones((num_sinks)), TimeMap[:, :, : (num_samples - wake_index), 1]
                ),
            )
            Ez[wake_index:] = np.dot(
                np.ones((num_sources)),
                np.dot(
                    np.ones((num_sinks)), TimeMap[:, :, : (num_samples - wake_index), 2]
                ),
            )

        # convert to etheta,ephi

    else:
        # create efiles for model
        if multiE:
            Ex = np.zeros(
                (desired_E_axis.shape[1], num_sinks, num_samples), dtype=np.float64
            )
            Ey = np.zeros(
                (desired_E_axis.shape[1], num_sinks, num_samples), dtype=np.float64
            )
            Ez = np.zeros(
                (desired_E_axis.shape[1], num_sinks, num_samples), dtype=np.float64
            )
            for e_inc in range(desired_E_axis.shape[1]):
                conformal_E_vectors = EM.calculate_conformalVectors(
                    desired_E_axis[e_inc, :],
                    np.asarray(aperture_coords.normals).astype(np.float32),antenna_axes
                )
                for element in range(num_sources):
                    point_informationv2[0:num_sources]["ex"] = 0.0
                    point_informationv2[0:num_sources]["ey"] = 0.0
                    point_informationv2[0:num_sources]["ez"] = 0.0
                    point_informationv2[element]["ex"] = (
                        conformal_E_vectors[element, 0] / num_sources
                    )
                    point_informationv2[element]["ey"] = (
                        conformal_E_vectors[element, 1] / num_sources
                    )
                    point_informationv2[element]["ez"] = (
                        conformal_E_vectors[element, 2] / num_sources
                    )
                    unified_weights[0:num_sources, :] = 0.0
                    unified_weights[element, :] = (
                        conformal_E_vectors[element, :] / num_sources
                    )
                    scattering_coefficient = 1 / 4 * scipy.constants.pi
                    TimeMap, WakeTimes = EM.TimeDomainv3(
                        num_sources,
                        num_sinks,
                        point_informationv2,
                        full_index,
                        scattering_coefficient,
                        wavelength,
                        excitation_function,
                        sampling_freq,
                        num_samples,
                    )
                    wake_index = np.digitize(WakeTimes, time_index)
                    Ex[element, wake_index:] = np.dot(
                        np.ones((num_sources)),
                        np.dot(
                            np.ones((num_sinks)),
                            TimeMap[:, :, : (num_samples - wake_index), 0],
                        ),
                    )
                    Ey[element, wake_index:] = np.dot(
                        np.ones((num_sources)),
                        np.dot(
                            np.ones((num_sinks)),
                            TimeMap[:, :, : (num_samples - wake_index), 1],
                        ),
                    )
                    Ez[element, wake_index:] = np.dot(
                        np.ones((num_sources)),
                        np.dot(
                            np.ones((num_sinks)),
                            TimeMap[:, :, : (num_samples - wake_index), 2],
                        ),
                    )

        else:
            Ex = np.zeros((num_sources, num_sinks, num_samples), dtype=np.float64)
            Ey = np.zeros((num_sources, num_sinks, num_samples), dtype=np.float64)
            Ez = np.zeros((num_sources, num_sinks, num_samples), dtype=np.float64)
            for element in range(num_sources):
                point_informationv2[0:num_sources]["ex"] = 0.0
                point_informationv2[0:num_sources]["ey"] = 0.0
                point_informationv2[0:num_sources]["ez"] = 0.0
                point_informationv2[element]["ex"] = (
                    conformal_E_vectors[element, 0] / num_sources
                )
                point_informationv2[element]["ey"] = (
                    conformal_E_vectors[element, 1] / num_sources
                )
                point_informationv2[element]["ez"] = (
                    conformal_E_vectors[element, 2] / num_sources
                )
                unified_weights[0:num_sources, :] = 0.0
                unified_weights[element, :] = (
                    conformal_E_vectors[element, :] / num_sources
                )
                scattering_coefficient = 1 / 4 * scipy.constants.pi
                TimeMap, WakeTimes = EM.TimeDomainv3(
                    num_sources,
                    num_sinks,
                    point_informationv2,
                    full_index,
                    scattering_coefficient,
                    wavelength,
                    excitation_function,
                    sampling_freq,
                    num_samples,
                )
                Ex[element, :, :] = np.dot(
                    np.ones((num_sources)),
                    np.dot(np.ones((num_sinks)), TimeMap[:, :, :, 0]),
                )
                Ey[element, :, :] = np.dot(
                    np.ones((num_sources)),
                    np.dot(np.ones((num_sinks)), TimeMap[:, :, :, 1]),
                )
                Ez[element, :, :] = np.dot(
                    np.ones((num_sources)),
                    np.dot(np.ones((num_sinks)), TimeMap[:, :, :, 2]),
                )

    return Ex, Ey, Ez, WakeTimes



import pytest
import numpy as np
import open3d as o3d
from numpy.testing import assert_equal, assert_allclose
from scipy.spatial.transform import Rotation as R
from ..electromagnetics.empropagation import vector_mapping

def test_vector_mapping_x_u():
    #initially test in global coordinate set
    rotation_matrix=R.from_euler('z',0,degrees=True).as_matrix()
    normal_vector=np.array([1.0, 0, 0])
    desired_E_vector=np.array([1.0, 0, 0],dtype=np.complex64) #uvn axes with n being the normal vector
    final_E_vector=np.array([0.0, 1.0, 0],dtype=np.complex64)
    assert_allclose(vector_mapping(desired_E_vector, normal_vector, rotation_matrix), final_E_vector,
                    atol=1e-12)

def test_vector_mapping_x_v():
    #initially test in global coordinate set
    rotation_matrix=R.from_euler('z',0,degrees=True).as_matrix()
    normal_vector=np.array([1.0, 0, 0])
    desired_E_vector=np.array([0.0, 1.0, 0],dtype=np.complex64) #uvn axes with n being the normal vector
    final_E_vector=np.array([0.0, 0.0, 1.0],dtype=np.complex64)
    assert_allclose(vector_mapping(desired_E_vector, normal_vector, rotation_matrix), final_E_vector,
                    atol=1e-12)

def test_vector_mapping_x_n():
    #initially test in global coordinate set
    rotation_matrix=R.from_euler('z',0,degrees=True).as_matrix()
    normal_vector=np.array([1.0, 0, 0])
    desired_E_vector=np.array([0, 0, 1.0],dtype=np.complex64) #uvn axes with n being the normal vector
    final_E_vector=np.array([1.0, 0, 0],dtype=np.complex64)
    assert_allclose(vector_mapping(desired_E_vector, normal_vector, rotation_matrix), final_E_vector,
                    atol=1e-12)

def test_vector_mapping_x_y_u():
    #initially test in global coordinate set
    rotation_matrix=R.from_euler('z',-90,degrees=True).as_matrix()
    normal_vector=np.array([1.0, 0, 0])
    desired_E_vector=np.array([1.0, 0, 0],dtype=np.complex64) #uvn axes with n being the normal vector
    final_E_vector=np.array([-1.0, 0.0, 0],dtype=np.complex64)
    assert_allclose(vector_mapping(desired_E_vector, normal_vector, rotation_matrix), final_E_vector,
                    atol=1e-12)

def test_vector_mapping_x_y_v():
    #initially test in global coordinate set
    rotation_matrix=R.from_euler('z',-90,degrees=True).as_matrix()
    normal_vector=np.array([1.0, 0, 0])
    desired_E_vector=np.array([0.0, 1.0, 0],dtype=np.complex64) #uvn axes with n being the normal vector
    final_E_vector=np.array([0.0, 0.0, 1.0],dtype=np.complex64)
    assert_allclose(vector_mapping(desired_E_vector, normal_vector, rotation_matrix), final_E_vector,
                    atol=1e-12)

def test_vector_mapping_x_y_n():
    #initially test in global coordinate set
    rotation_matrix=R.from_euler('z',-90,degrees=True).as_matrix()
    normal_vector=np.array([1.0, 0, 0])
    desired_E_vector=np.array([0, 0, 1.0],dtype=np.complex64) #uvn axes with n being the normal vector
    final_E_vector=np.array([0.0, 1.0, 0],dtype=np.complex64)
    assert_allclose(vector_mapping(desired_E_vector,normal_vector,rotation_matrix), final_E_vector,
                    atol=1e-12)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
import math
import numpy as np
import open3d as o3d
import scipy.io as io
from importlib_resources import files

import lyceanem.tests.data
import lyceanem.geometry.targets as tl
import lyceanem.geometry.geometryfunctions as GF
import lyceanem.raycasting.rayfunctions as RF


# freq=24e9
# wavelength=3e8/freq
# measurementaddress=Path("./")
# units are mm in the laser scans
# medium_reference='Medium_Reference_Plate_Covered_Normal_1_plan_grid_0p5mm.txt'
# matfile='plate_copper.mat'
# datafile='Medium_Reference_Plate_Covered_Normal_1_plan_grid_0p5mm.txt'
# stream=pkg_resources.resource_stream(__name__,'data/Medium_Reference_Plate_Covered_Normal_1_plan_grid_0p5mm.txt')
# temp=np.loadtxt(stream,delimiter=';',skiprows=2)

# medium_reference_reflector = o3d.geometry.PointCloud()
# medium_reference_reflector.points = o3d.utility.Vector3dVector(temp/1000)
# medium_reference_reflector.estimate_normals()
# reference_point=np.asarray([[np.min(temp[:,0]),np.min(temp[:,1]),np.min(temp[:,2])]])/1000

# downsampled_reflector=medium_reference_reflector.voxel_down_sample(voxel_size=wavelength*0.5)
# downsampled_reflector.translate(-reference_point.ravel(),relative=True)
# downsampled_reflector.translate([-0.15,-0.15,0],relative=True)
def exampleUAV(frequency):
    bodystream = files(lyceanem.tests.data).joinpath("UAV.stl")
    arraystream = files(lyceanem.tests.data).joinpath("UAVarray.stl")
    body = o3d.io.read_triangle_mesh(str(bodystream))
    array = o3d.io.read_triangle_mesh(str(arraystream))
    rotation_vector1 = np.asarray([0.0, np.deg2rad(90), 0.0])
    rotation_vector2 = np.asarray([np.deg2rad(90), 0.0, 0.0])
    body = GF.open3drotate(
        body, o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector1)
    )
    body = GF.open3drotate(
        body, o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector2)
    )
    array = GF.open3drotate(
        array, o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector1)
    )
    array = GF.open3drotate(
        array, o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector2)
    )
    body.translate(np.asarray([0.25, 0, 0]), relative=True)
    array.translate(np.asarray([0.25, 0, 0]), relative=True)
    array.translate(np.array([-0.18, 0, 0.0125]), relative=True)
    body.translate(np.array([-0.18, 0, 0.0125]), relative=True)
    array.compute_vertex_normals()
    array.paint_uniform_color(np.array([0, 1.0, 1.0]))
    body.compute_vertex_normals()
    body.paint_uniform_color(np.array([0, 0.259, 0.145]))

    wavelength = 3e8 / frequency
    mesh_sep = wavelength * 0.5

    # define UAV array, a line, a circle, then a line with reference to the array solid
    array_vertices = np.asarray(array.vertices)
    array_height = np.max(array_vertices[:, 2]) - np.min(array_vertices[:, 2])
    top_index = (
        array_vertices[:, 2] >= np.max(array_vertices[:, 2]) - 1e-4
    )  # pick only vertices on the top of the array
    lower_index = (
        array_vertices[:, 2] <= np.min(array_vertices[:, 2]) + 1e-4
    )  # pick only vertices on the bottom of the surface
    array_top_points = array_vertices[top_index, :]
    array_bottom_points = array_vertices[lower_index, :]
    r1 = np.max(array_bottom_points[:, 1])
    r2 = np.mean(
        np.asarray([np.max(array_vertices[:, 1]), np.abs(np.min(array_vertices[:, 1]))])
    )
    top_length = np.abs(np.min(array_top_points[:, 0]) - np.max(array_top_points[:, 0]))
    bottom_length = np.abs(
        np.min(array_bottom_points[:, 0]) - np.max(array_bottom_points[:, 0])
    )
    slant_adjustment = (top_length - r2) - (bottom_length - r1)
    # r2 center is 25mm further forward than r1
    l = bottom_length - r1

    slant_angle = np.arctan2(r2 - r1, array_height)
    slant_h = (array_height ** 2 + (r2 - r1) ** 2) ** 0.5
    rows = math.ceil((slant_h) / (mesh_sep) + 1)
    heights = np.linspace(0, slant_h, rows)
    centers = np.linspace(l + 0, l + slant_adjustment, rows)
    rad_sep = np.linspace(r1, r2, rows)
    nose_distances = np.linspace(
        0, rad_sep[0] * np.pi, math.ceil((rad_sep[0] * np.pi) / (mesh_sep) + 1)
    )
    nose_side1 = np.empty((0, 3), dtype=np.float32)
    nose_side2 = np.empty((0, 3), dtype=np.float32)
    nose_circ = np.empty((0, 3), dtype=np.float32)
    nose_circ_normals = np.empty((0, 3), dtype=np.float32)
    for nose_inc in range(rows):
        nose_x = np.linspace(
            l, centers[nose_inc], math.ceil(((centers[nose_inc] - l) / mesh_sep + 1))
        )[1:]
        nose_y = (
            np.zeros((len(nose_x.ravel())), dtype=np.float32)
            + heights[nose_inc] * np.sin(slant_angle)
            + r1
        )
        nose_z = np.zeros((len(nose_x.ravel())), dtype=np.float32) + heights[
            nose_inc
        ] * np.cos(slant_angle)
        nose_side1 = np.append(
            nose_side1, np.array([nose_x, nose_y, nose_z]).transpose(), axis=0
        )
        nose_side2 = np.append(
            nose_side2, np.array([nose_x, -nose_y, nose_z]).transpose(), axis=0
        )
        angle_inc = np.pi
        nose_angles = np.linspace(
            0, np.pi, math.ceil((rad_sep[nose_inc] * np.pi) / (mesh_sep) + 1)
        )
        tempx = rad_sep[nose_inc] * np.sin(nose_angles)[1:-1] + centers[nose_inc]
        tempy = rad_sep[nose_inc] * np.cos(nose_angles)[1:-1]
        tempz = np.zeros((len(tempx.ravel())), dtype=np.float32) + heights[
            nose_inc
        ] * np.cos(slant_angle)
        testnose = np.array([tempx, tempy, tempz]).transpose()
        nose_circ = np.append(nose_circ, testnose, axis=0)
        tempnormx = np.cos(nose_angles[1:-1] - np.pi / 2)
        tempnormy = -np.sin(nose_angles[1:-1] - np.pi / 2)
        tempnormz = np.ones((len(nose_angles[1:-1])), dtype=np.float32) * np.sin(
            slant_angle + np.pi / 2 + np.pi / 2
        )
        nose_circ_normals = np.append(
            nose_circ_normals,
            np.array([tempnormx, tempnormy, tempnormz]).transpose(),
            axis=0,
        )

    nose_side1_normals = np.zeros((len(nose_side1), 3), dtype=np.float32)
    nose_side2_normals = np.zeros((len(nose_side2), 3), dtype=np.float32)
    nose_side2_normals[:, 1] = np.cos(slant_angle + np.pi / 2 + np.pi / 2)
    nose_side2_normals[:, 2] = np.sin(slant_angle + np.pi / 2 + np.pi / 2)
    nose_side1_normals[:, 1] = np.cos(np.pi / 2 - slant_angle - np.pi / 2)
    nose_side1_normals[:, 2] = np.sin(np.pi / 2 - slant_angle - np.pi / 2)

    side1x = np.linspace(0, l, math.ceil((l) / (mesh_sep) + 1))
    side1y = heights * np.sin(-slant_angle)
    side1z = heights * np.cos(slant_angle)
    xx, yy = np.meshgrid(side1x, heights)
    side1 = np.zeros(((len(np.ravel(xx)), 3)), dtype=np.float32)
    side2_normals = np.zeros(((len(np.ravel(xx)), 3)), dtype=np.float32)
    side2_normals[:, 1] = np.cos(slant_angle + np.pi / 2 + np.pi / 2)
    side2_normals[:, 2] = np.sin(slant_angle + np.pi / 2 + np.pi / 2)
    side1_normals = np.zeros(((len(np.ravel(xx)), 3)), dtype=np.float32)
    side1_normals[:, 1] = np.cos(np.pi / 2 - slant_angle - np.pi / 2)
    side1_normals[:, 2] = np.sin(np.pi / 2 - slant_angle - np.pi / 2)
    side1[:, 0] = xx.ravel()
    side1[:, 1] = yy.ravel() * np.sin(slant_angle) + r1
    side1[:, 2] = yy.ravel() * np.cos(slant_angle)
    side2 = np.empty(((len(np.ravel(xx)), 3)), dtype=np.float32)
    side2[:, 0] = xx.ravel()
    side2[:, 1] = yy.ravel() * np.sin(-slant_angle) - r1
    side2[:, 2] = yy.ravel() * np.cos(-slant_angle)

    total_array = np.append(
        np.append(
            np.append(np.append(side1, side2, axis=0), nose_circ, axis=0),
            nose_side1,
            axis=0,
        ),
        nose_side2,
        axis=0,
    ) + np.array([0.1025, 0, -0.025])
    total_array_normals = np.append(
        np.append(
            np.append(
                np.append(side1_normals, side2_normals, axis=0),
                nose_circ_normals,
                axis=0,
            ),
            nose_side1_normals,
            axis=0,
        ),
        nose_side2_normals,
        axis=0,
    ) + np.array([0.1025, 0, -0.025])
    source_pcd = RF.points2pointcloud(total_array)
    source_pcd.normals = o3d.utility.Vector3dVector(total_array_normals)
    source_pcd.translate(np.array([-0.18, 0, 0.0125]), relative=True)

    return body, array, source_pcd


def prototypescan(mesh_size, wavelength=3e8 / 24e9, importpcd=True):
    if importpcd:
        if mesh_size < (wavelength * 0.5):
            # stream = pkg_resources.resource_stream(__name__,
            #                                       'data/prototypescanpoints24GHztenthwavelength.ply')
            stream = files("lyceanem.tests.data").joinpath(
                "prototypescanpoints24GHztenthwavelength.ply"
            )
            prototype_points = o3d.io.read_point_cloud(str(stream))
        else:
            # stream = pkg_resources.resource_stream(__name__,
            #                                       'data/prototypescanpoints24GHztenthwavelength.ply')
            stream = files(lyceanem.tests.data).joinpath(
                "prototypescanpoints24GHztenthwavelength.ply"
            )
            prototype_points = o3d.io.read_point_cloud(str(stream))
        prototype_mesh, _ = tl.meshedReflector(0.3, 0.3, 6e-3, mesh_size)
    else:
        # stream = pkg_resources.resource_stream(__name__,
        #                                       'data/plate_copper.mat')
        stream = files("lyceanem.tests.data").joinpath("plate_copper.mat")
        matdata = io.loadmat(str(stream))["Flat_prototype"]
        copper_reference_point = (
            np.asarray(
                [[np.min(matdata[:, 0]), np.min(matdata[:, 1]), np.min(matdata[:, 2])]]
            )
            / 1000
        )
        prototype = o3d.geometry.PointCloud()
        # convert from mm to si units
        prototype.points = o3d.utility.Vector3dVector(matdata / 1000)
        prototype.estimate_normals()
        prototype.orient_normals_to_align_with_direction(
            orientation_reference=[0.0, 0.0, 1.0]
        )
        downsampled_prototype = prototype.voxel_down_sample(voxel_size=mesh_size * 1.1)
        downsampled_prototype.translate(-copper_reference_point.ravel(), relative=True)
        prototype.translate(-copper_reference_point.ravel(), relative=True)
        downsampled_prototype.translate([-0.15, -0.15, 0], relative=True)
        prototype.translate([-0.15, -0.15, 0], relative=True)
        reflectorplate, scatter_points = tl.meshedReflector(
            0.3, 0.3, 6e-3, mesh_size, sides="all"
        )
        temp_points = np.asarray(scatter_points.points)
        temp_points_back = temp_points[temp_points[:, 2] < -1e-3, :]
        temp_points_top = temp_points[temp_points[:, 1] > 1.49e-1, :]
        top_normals = np.zeros((temp_points_top.shape[0], 3))
        top_normals[:, 1] = 1
        toppoints = o3d.geometry.PointCloud()
        toppoints.points = o3d.utility.Vector3dVector(temp_points_top)
        toppoints.normals = o3d.utility.Vector3dVector(top_normals)
        sideapoints = copy.deepcopy(toppoints)
        rotation_vector1 = np.radians(np.array([0, 0, 90]))
        sideapoints = GF.open3drotate(
            sideapoints,
            o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector1),
        )
        bottompoints = copy.deepcopy(sideapoints)
        bottompoints = GF.open3drotate(
            bottompoints,
            o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector1),
        )
        sidebpoints = copy.deepcopy(bottompoints)
        sidebpoints = GF.open3drotate(
            sidebpoints,
            o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector1),
        )

        back_normals = np.zeros((temp_points_back.shape[0], 3))
        back_normals[:, 2] = -1
        backpoints = o3d.geometry.PointCloud()
        backpoints.points = o3d.utility.Vector3dVector(temp_points_back)
        backpoints.normals = o3d.utility.Vector3dVector(back_normals)
        # for now only interested in points on the front surface
        temp = np.asarray(downsampled_prototype.points)
        interest_reference = np.min(temp[:, 2])
        test_prototype = (
            downsampled_prototype
            + backpoints
            + toppoints
            + bottompoints
            + sideapoints
            + sidebpoints
        )
        (
            prototype_mesh,
            output,
        ) = o3d.geometry.TriangleMesh.create_from_point_cloud_poisson(
            test_prototype, linear_fit=True
        )
        prototype_mesh.compute_triangle_normals()
        prototype_points = o3d.geometry.PointCloud()
        vertices = np.asarray(prototype_mesh.vertices)
        normals = np.asarray(prototype_mesh.vertex_normals)
        prototype_points.points = o3d.utility.Vector3dVector(
            vertices[vertices[:, 2] >= interest_reference, :]
        )
        prototype_points.normals = o3d.utility.Vector3dVector(
            normals[vertices[:, 2] >= interest_reference, :]
        )
    return prototype_mesh, prototype_points


def mediumreferencescan(mesh_size, wavelength=3e8 / 24e9, importpcd=True):
    if importpcd:
        if mesh_size < (wavelength * 0.5):
            # stream = pkg_resources.resource_stream(__name__,
            #                                       'data/referencescanpoints24GHztenthwavelength.ply')
            stream = files("lyceanem.tests.data").joinpath(
                "referencescanpoints24GHztenthwavelength.ply"
            )
            reflector_points = o3d.io.read_point_cloud(str(stream))
        else:
            # stream = pkg_resources.resource_stream(__name__,
            #                                       'data/referencescanpoints24GHztenthwavelength.ply')
            stream = files("lyceanem.tests.data").joinpath(
                "referencescanpoints24GHztenthwavelength.ply"
            )
            reflector_points = o3d.io.read_point_cloud(str(stream))
        reference_mesh, _ = tl.meshedReflector(0.3, 0.3, 6e-3, mesh_size)
    else:
        # stream=pkg_resources.resource_stream(__name__,'data/Medium_Reference_Plate_Covered_Normal_1_plan_grid_0p5mm.txt')
        stream = files("lyceanem.tests.data").joinpath(
            "Medium_Reference_Plate_Covered_Normal_1_plan_grid_0p5mm.txt"
        )
        temp = np.loadtxt(str(stream), delimiter=";", skiprows=2)
        medium_reference_reflector = o3d.geometry.PointCloud()
        medium_reference_reflector.points = o3d.utility.Vector3dVector(temp / 1000)
        medium_reference_reflector.estimate_normals()
        medium_reference_reflector.orient_normals_to_align_with_direction(
            orientation_reference=[0.0, 0.0, 1.0]
        )
        reference_point = (
            np.asarray([[np.min(temp[:, 0]), np.min(temp[:, 1]), np.min(temp[:, 2])]])
            / 1000
        )

        downsampled_reflector = medium_reference_reflector.voxel_down_sample(
            voxel_size=mesh_size * 1.1
        )
        downsampled_reflector.translate(-reference_point.ravel(), relative=True)
        medium_reference_reflector.translate(-reference_point.ravel(), relative=True)
        downsampled_reflector.translate([-0.15, -0.15, 0], relative=True)
        medium_reference_reflector.translate([-0.15, -0.15, 0], relative=True)
        reflectorplate, scatter_points = tl.meshedReflector(
            0.3, 0.3, 6e-3, mesh_size, sides="all"
        )
        temp_points = np.asarray(scatter_points.points)
        temp_points_back = temp_points[temp_points[:, 2] < -1e-3, :]
        temp_points_top = temp_points[temp_points[:, 1] > 1.49e-1, :]
        top_normals = np.zeros((temp_points_top.shape[0], 3))
        top_normals[:, 1] = 1
        toppoints = o3d.geometry.PointCloud()
        toppoints.points = o3d.utility.Vector3dVector(temp_points_top)
        toppoints.normals = o3d.utility.Vector3dVector(top_normals)
        sideapoints = copy.deepcopy(toppoints)
        rotation_vector1 = np.radians(np.array([0, 0, 90]))
        sideapoints = GF.open3drotate(
            sideapoints,
            o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector1),
        )

        bottompoints = copy.deepcopy(sideapoints)
        bottompoints = GF.open3drotate(
            bottompoints,
            o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector1),
        )

        sidebpoints = copy.deepcopy(bottompoints)
        sidebpoints = GF.open3drotate(
            sidebpoints,
            o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector1),
        )

        back_normals = np.zeros((temp_points_back.shape[0], 3))
        back_normals[:, 2] = -1
        backpoints = o3d.geometry.PointCloud()
        backpoints.points = o3d.utility.Vector3dVector(temp_points_back)
        backpoints.normals = o3d.utility.Vector3dVector(back_normals)
        temp = np.asarray(downsampled_reflector.points)
        interest_reference = np.min(temp[:, 2])
        test_prototype = (
            downsampled_reflector
            + backpoints
            + toppoints
            + bottompoints
            + sideapoints
            + sidebpoints
        )
        (
            reference_mesh,
            output,
        ) = o3d.geometry.TriangleMesh.create_from_point_cloud_poisson(
            test_prototype, linear_fit=True
        )
        reference_mesh.compute_triangle_normals()
        reflector_points = o3d.geometry.PointCloud()
        vertices = np.asarray(reference_mesh.vertices)
        normals = np.asarray(reference_mesh.vertex_normals)
        reflector_points.points = o3d.utility.Vector3dVector(
            vertices[vertices[:, 2] >= interest_reference, :]
        )
        reflector_points.normals = o3d.utility.Vector3dVector(
            normals[vertices[:, 2] >= interest_reference, :]
        )
    return reference_mesh, reflector_points



import copy

import numpy as np
import open3d as o3d
from numba import vectorize
from packaging import version
from scipy.spatial.transform import Rotation as R
from .. import base_types as base_types
from .. import base_classes as base_classes
from ..raycasting import rayfunctions as RF


def tri_areas(solid):
    """
    Takes as an import an open3D triangle mesh, then calculates the area of all triangles in the global units, and
    returns as an array of areas
    """
    triangle_vertices = np.asarray(solid.vertices)
    triangleidx = np.asarray(solid.triangles)
    a_vectors = (
        triangle_vertices[triangleidx[:, 1], :]
        - triangle_vertices[triangleidx[:, 0], :]
    )
    b_vectors = (
        triangle_vertices[triangleidx[:, 2], :]
        - triangle_vertices[triangleidx[:, 0], :]
    )
    u = np.cross(a_vectors, b_vectors)
    areas = 0.5 * ((u[:, 0] ** 2 + u[:, 1] ** 2 + u[:, 2] ** 2) ** 0.5)

    return areas


def tri_centroids(solid):
    """
    In order to calculate the centroid of the triangle, take vertices from open3d triangle mesh, then put each triangle
    into origin space, creating oa,ob,oc vectors, then the centroid is a third of the sum of oa+ob+oc, converted back
    into global coordinates
    """
    triangle_vertices = np.asarray(solid.vertices)
    triangleidx = np.asarray(solid.triangles)
    oa = triangle_vertices[triangleidx[:, 0], :]
    ob = triangle_vertices[triangleidx[:, 1], :]
    oc = triangle_vertices[triangleidx[:, 2], :]
    centroids = (1 / 3) * (oa + ob + oc)
    centroid_cloud = o3d.geometry.PointCloud()
    centroid_cloud.points = o3d.utility.Vector3dVector(centroids)
    centroid_cloud.normals = solid.triangle_normals
    return centroids, centroid_cloud


def decimate_mesh(solid, mesh_sep):
    """
    In order to calculate the scattering appropriately the triangle mesh should be decimated so that the vertices
    are spaced mesh_sep apart.
    inputs are the :class:`open3d.geometry.TriangleMesh` solid, and the mesh_sep, and the output is a new :class:`open3d.geometry.TriangleMesh`. This is only required for
    the discrete scattering model, using the centroids or vertices
    """
    new_solid = o3d.geometry.TriangleMesh()
    # lineset=o3d.geometry.LineSet.create_from_triangle_mesh(solid)
    # identify triangles which are too large via areas greater than mesh_sep**2
    area_limit = mesh_sep ** 2
    areas = tri_areas(solid)
    large_tri_index = np.where(areas > area_limit)[0]
    while np.max(areas)>area_limit:
        solid=solid.subdivide_midpoint()
        areas = tri_areas(solid)

    new_solid=copy.deepcopy(solid)
    return new_solid


@vectorize(["(float32(float32))", "(float64(float64))"])
def elevationtotheta(el):
    # converting elevation in degrees to theta in degrees
    # elevation is in range -90 to 90 degrees
    # theta is in range 0 to 180 degrees
    if el >= 0.0:
        theta = 90.0 - el
    else:
        theta = np.abs(el) + 90.0

    return theta


@vectorize(["(float32(float32))", "(float64(float64))"])
def thetatoelevation(theta):
    # converting theta in degrees to elevation in degrees
    # elevation is in range -90 to 90 degrees
    # theta is in range 0 to 180 degrees
    if theta <= 90.0:
        # theta=(90.0-el)
        el = 90 - theta
    else:
        # theta=np.abs(el)+90.0
        el = -(theta - 90.0)

    return el


def open3drotate(
    item, rotation_matrix, rotation_centre=np.zeros((3, 1), dtype=np.float32)
):
    """

    Parameters
    ----------
    item : open3d object
        :class:`open3d.geometry.PointCloud`, :class:`open3d.geometry.TriangleMesh` or other open3d objects with .rotate function
    rotation_matrix : open3d rotation matrix
        rotation matrix for the desired transformation
    rotation_centre : numpy float (3,1)
        desired rotation centre, defaults to the origin

    Returns
    -------
    item : open3d object
        rotated item
    """
    if version.parse(o3d.__version__) >= version.parse("0.10.0"):
        # new syntax for rotations
        item.rotate(rotation_matrix, center=rotation_centre)
    else:

        item.translate(-1 * rotation_centre)
        item.rotate(rotation_matrix, center=False)
        item.translate(rotation_centre)

    return item


def get_cross_prod_mat(pVec_Arr):
    # pVec_Arr shape (3)
    qCross_prod_mat = np.array(
        [
            [0, -pVec_Arr[2], pVec_Arr[1]],
            [pVec_Arr[2], 0, -pVec_Arr[0]],
            [-pVec_Arr[1], pVec_Arr[0], 0],
        ]
    )
    return qCross_prod_mat


def calculate_align_mat(pVec_Arr):
    scale = np.linalg.norm(pVec_Arr)
    pVec_Arr = pVec_Arr / scale
    # must ensure pVec_Arr is also a unit vec.
    z_unit_Arr = np.array([0, 0, 1])
    z_mat = get_cross_prod_mat(z_unit_Arr)

    z_c_vec = np.matmul(z_mat, pVec_Arr)
    z_c_vec_mat = get_cross_prod_mat(z_c_vec)

    qTrans_Mat = (
        np.eye(3, 3)
        + z_c_vec_mat
        + np.matmul(z_c_vec_mat, z_c_vec_mat) / (1 + np.dot(z_unit_Arr, pVec_Arr))
    )
    qTrans_Mat *= scale
    return qTrans_Mat


def axes_from_normal(boresight_vector, boresight_along="x"):
    """
    Calculates the required local axes within the global coordinate frame based upon the desired boresight, the standard
    is that the x-axis will be pointed along the desired boresight_vector, while the z-axis will be aligned as closely
    as possible to the global z-axis.

    Parameters
    ----------
    boresight_vector

    Returns
    -------
    rotation_matrix

    """
    # initially define the rotation matrix based inline with the global coordinate frame.
    #for more consistency, need to calculate additional vectors to ensure rotation mapping is correct.

    replacement_vector = boresight_vector / np.linalg.norm(boresight_vector)

    if boresight_along == "x":
        alignment_vector=np.array([[1.0,0,0]])
        u_vector=np.array([[0.0,1.0,0]])
        v_vector=np.array([[0,0,1.0]])
    elif boresight_along == "y":
        alignment_vector=np.array([[0.0,1.0,0]])
        u_vector = np.array([[1.0, 0.0, 0]])
        v_vector = np.array([[0, 0, 1.0]])
    elif boresight_along == "z":
        alignment_vector=np.array([[0.0,0,1.0]])
        u_vector = np.array([[1.0, 0.0, 0]])
        v_vector = np.array([[0, 1.0, 0]])
    rotation,_=R.align_vectors(replacement_vector.reshape(1,3),alignment_vector)
    # if boresight_along == "x":
    #     # then u should be based on y, and v on z
    #     rotation_matrix[1, :] = np.cross(
    #         rotation_matrix[2, :], rotation_matrix[0, :]
    #     ) / np.linalg.norm(np.cross(rotation_matrix[2, :], rotation_matrix[0, :]))
    #     rotation_matrix[2, :] = np.cross(
    #         rotation_matrix[0, :], rotation_matrix[1, :]
    #     ) / np.linalg.norm(np.cross(rotation_matrix[0, :], rotation_matrix[1, :]))
    #     if np.any(np.isnan(rotation_matrix)):
    #         # nan in rotation_matrix, start again with different basis vector
    #         rotation_matrix = np.eye(3)
    #         rotation_matrix[0, :] = replacement_vector
    #         rotation_matrix[2, :] = np.cross(
    #             rotation_matrix[0, :], rotation_matrix[1, :]
    #         ) / np.linalg.norm(np.cross(rotation_matrix[0, :], rotation_matrix[1, :]))
    #         rotation_matrix[1, :] = np.cross(
    #             rotation_matrix[2, :], rotation_matrix[0, :]
    #         ) / np.linalg.norm(np.cross(rotation_matrix[2, :], rotation_matrix[0, :]))
    # elif boresight_along == "y":
    #     # then u should be based on x, and v on z
    #     rotation_matrix[0, :] = np.cross(
    #         rotation_matrix[1, :], rotation_matrix[2, :]
    #     ) / np.linalg.norm(np.cross(rotation_matrix[1, :], rotation_matrix[2, :]))
    #     rotation_matrix[2, :] = np.cross(
    #         rotation_matrix[0, :], rotation_matrix[1, :]
    #     ) / np.linalg.norm(np.cross(rotation_matrix[0, :], rotation_matrix[1, :]))
    #     if np.any(np.isnan(rotation_matrix)):
    #         # nan in rotation_matrix, start again with different basis vector
    #         rotation_matrix = np.eye(3)
    #         rotation_matrix[1, :] = replacement_vector
    #         rotation_matrix[2, :] = np.cross(
    #             rotation_matrix[0, :], rotation_matrix[1, :]
    #         ) / np.linalg.norm(np.cross(rotation_matrix[0, :], rotation_matrix[1, :]))
    #         rotation_matrix[0, :] = np.cross(
    #             rotation_matrix[1, :], rotation_matrix[2, :]
    #         ) / np.linalg.norm(np.cross(rotation_matrix[1, :], rotation_matrix[2, :]))
    # elif boresight_along == "z":
    #     # then u should be based on x, and v on y
    #     rotation_matrix[0, :] = np.cross(
    #         rotation_matrix[1, :], rotation_matrix[2, :]
    #     ) / np.linalg.norm(np.cross(rotation_matrix[1, :], rotation_matrix[2, :]))
    #     rotation_matrix[1, :] = np.cross(
    #         rotation_matrix[2, :], rotation_matrix[0, :]
    #     ) / np.linalg.norm(np.cross(rotation_matrix[2, :], rotation_matrix[0, :]))
    #     if np.any(np.isnan(rotation_matrix)):
    #         # nan in rotation_matrix, start again with different basis vector
    #         print('nan')
    #         rotation_matrix = np.eye(3)
    #         rotation_matrix[2, :] = replacement_vector
    #         rotation_matrix[1, :] = np.cross(
    #             rotation_matrix[2, :], rotation_matrix[0, :]
    #         ) / np.linalg.norm(np.cross(rotation_matrix[2, :], rotation_matrix[0, :]))
    #         rotation_matrix[0, :] = np.cross(
    #             rotation_matrix[1, :], rotation_matrix[2, :]
    #         ) / np.linalg.norm(np.cross(rotation_matrix[1, :], rotation_matrix[2, :]))

    return rotation.as_matrix()

def mesh_conversion(conversion_object):
    """
    Convert the provide file object into triangle_t format

    Parameters
    ----------
    conversion_object : solid object to be converted into triangle_t format, could be open3d trianglemesh, solid, or antenna structure

    Returns
    -------
    triangles : numpy array of type triangle_t
    """
    if isinstance(conversion_object,base_classes.structures):
        triangles = object.triangles_base_raycaster()
    elif isinstance(conversion_object,base_classes.antenna_structures):
        exported_structure=base_classes.structures(solids=object.export_all_structures())
        triangles=exported_structure.triangles_base_raycaster()
    elif isinstance(conversion_object,type(o3d.geometry.TriangleMesh())):
        triangles=RF.convertTriangles(object)
    elif isinstance(conversion_object,list):
        triangles = np.empty((0), dtype=base_types.triangle_t)
        #print("Detected List")
        for item in conversion_object:
            #print(type(item))
            #print(isinstance(item,type(o3d.geometry.TriangleMesh())))
            if isinstance(item, type(o3d.geometry.TriangleMesh())):
                triangles = np.append(triangles,RF.convertTriangles(item),axis=0)
                #print(len(triangles))
            elif isinstance(item, base_classes.structures):
                triangles=np.append(triangles,item.triangles_base_raycaster(),axis=0)
                #print(len("B ",triangles))

    else:
        print("no structures")
        print(type(object))
        triangles = np.empty((0), dtype=base_types.triangle_t)

    return triangles




import math
from packaging import version
import numpy as np
from numba import float32, from_dtype, njit, guvectorize

@njit
def cart2pol(x, y):
    rho = np.sqrt(x ** 2 + y ** 2)
    phi = np.arctan2(y, x)
    return (rho, phi)


@njit
def pol2cart(rho, phi):
    x = rho * np.cos(phi)
    y = rho * np.sin(phi)
    return (x, y)


@njit
def cart2sph(x, y, z):
    # radians
    hxy = np.hypot(x, y)
    r = np.hypot(hxy, z)
    el = np.arctan2(z, hxy)
    az = np.arctan2(y, x)
    return az, el, r


@njit
def sph2cart(az, el, r):
    # radians
    rcos_theta = r * np.cos(el)
    x = rcos_theta * np.cos(az)
    y = rcos_theta * np.sin(az)
    z = r * np.sin(el)
    return x, y, z


@njit
def calc_normals(T):
    # calculate triangle norm
    e1x = T["v1x"] - T["v0x"]
    e1y = T["v1y"] - T["v0y"]
    e1z = T["v1z"] - T["v0z"]

    e2x = T["v2x"] - T["v0x"]
    e2y = T["v2y"] - T["v0y"]
    e2z = T["v2z"] - T["v0z"]

    dirx = e1y * e2z - e1z * e2y
    diry = e1z * e2x - e1x * e2z
    dirz = e1x * e2y - e1y * e2x

    normconst = math.sqrt(dirx ** 2 + diry ** 2 + dirz ** 2)

    T["normx"] = dirx / normconst
    T["normy"] = diry / normconst
    T["normz"] = dirz / normconst

    return T


@guvectorize(
    [(float32[:], float32[:], float32[:], float32)],
    "(n),(n)->(n),()",
    target="parallel",
)
def fast_calc_dv(source, target, dv, normconst):
    dirx = target[0] - source[0]
    diry = target[1] - source[1]
    dirz = target[2] - source[2]
    normconst = math.sqrt(dirx ** 2 + diry ** 2 + dirz ** 2)
    dv = np.array([dirx, diry, dirz]) / normconst


@njit
def calc_dv(source, target):
    dirx = target[0] - source[0]
    diry = target[1] - source[1]
    dirz = target[2] - source[2]
    normconst = np.sqrt(dirx ** 2 + diry ** 2 + dirz ** 2)
    dv = np.array([dirx, diry, dirz]) / normconst
    return dv[0], dv[1], dv[2], normconst


@njit
def calc_dv_norm(source, target, direction, length):
    length[:, 0] = np.sqrt(
        (target[:, 0] - source[:, 0]) ** 2
        + (target[:, 1] - source[:, 1]) ** 2
        + (target[:, 2] - source[:, 2]) ** 2
    )
    direction = (target - source) / length
    return direction, length


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jun 16 13:15:07 2020

@author: timtitan
"""
import logging
from timeit import default_timer as timer

import numpy as np
import open3d as o3d
from tqdm import tqdm

import lyceanem.geometry.targets as TL
import lyceanem.models.frequency_domain as FD
import lyceanem.raycasting.rayfunctions as RF

# import GPUScatteringTest as GU
import lyceanem.tests.reflectordata as RD

nb_logger = logging.getLogger("numba")
nb_logger.setLevel(logging.ERROR)  # only show error
angle_values = np.linspace(0, 90, 91)
max_scatter = 2
responsex2 = np.zeros((len(angle_values), 3), dtype="complex")
responsey2 = np.zeros((len(angle_values), 3), dtype="complex")
responsez2 = np.zeros((len(angle_values), 3), dtype="complex")
isoresponse = np.zeros((len(angle_values)), dtype="complex")
# responsey3=np.zeros((len(angle_values),3),dtype='complex')
# responsez3=np.zeros((len(angle_values),3),dtype='complex')


freq = np.asarray(6.0e9)
wavelength = 3e8 / freq
mesh_resolution = 0.5 * wavelength
# point_area=(mesh_resolution)**2
# reflectorplate_main,_=TL.meshedReflector(0.3,0.3,6e-3,wavelength*0.5,sides='front')
# scatter_points_main=TL.meshsolid(reflectorplate_main, point_area)

resolution = 5
show_scene = False
# response2=np.zeros((len(angle_values),3),dtype='complex')
for angle_inc in tqdm(range(len(angle_values))):
    overalltime = timer()
    # define antenna surface
    # freq=np.asarray(24.0e9)
    wavelength = 3e8 / freq
    # mesh_resolution=0.5*wavelength
    aperture_resolution = 0.5 * wavelength
    single_source = False
    single_sink = False
    if single_source:
        x = np.full((1, 1), 0.0, dtype=np.float32)
        z = np.full((1, 1), 0.0, dtype=np.float32)
        xx, yy = np.meshgrid(x, z)
        zz = np.zeros((len(np.ravel(xx)), 1))
        transmitting_antenna_surface_coords = RF.points2pointcloud(np.zeros((1, 3)))
        normals = np.zeros((1, 3))
        normals[0, 2] = 1.0
        transmitting_antenna_surface_coords.normals = o3d.utility.Vector3dVector(
            normals
        )
        transmit_horn_structure, _ = TL.meshedHorn(
            58e-3, 58e-3, 128e-3, 2e-3, 0.21, aperture_resolution
        )
    else:
        # horn spec with x_width, y_width, length, edge width, flare angle, and aperture resolution
        transmit_horn_structure, transmitting_antenna_surface_coords = TL.meshedHorn(
            58e-3, 58e-3, 128e-3, 2e-3, 0.21, aperture_resolution
        )

    if single_sink:
        x = np.full((1, 1), 0.0, dtype=np.float32)
        z = np.full((1, 1), 0.0, dtype=np.float32)
        xx, yy = np.meshgrid(x, z)
        zz = np.zeros((len(np.ravel(xx)), 1))
        recieving_antenna_surface_coords = RF.points2pointcloud(np.zeros((1, 3)))
        normals = np.zeros((1, 3))
        normals[0, 2] = 1.0
        recieving_antenna_surface_coords.normals = o3d.utility.Vector3dVector(normals)
        # recieving_antenna_surface_coords=np.zeros(((len(np.ravel(xx)),3)),dtype=np.float32)
        # recieving_antenna_surface_coords[:,0]=np.ravel(xx)
        # recieving_antenna_surface_coords[:,1]=np.ravel(yy)
        # recieving_antenna_surface_coords[:,2]=np.ravel(zz)
        receive_horn_structure, _ = TL.meshedHorn(
            58e-3, 58e-3, 128e-3, 2e-3, 0.21, aperture_resolution
        )
    else:
        receive_horn_structure, recieving_antenna_surface_coords = TL.meshedHorn(
            58e-3, 58e-3, 128e-3, 2e-3, 0.21, aperture_resolution
        )

    # rx90=R.from_euler('x',90,degrees=True)
    # rz90=R.from_euler('z',90,degrees=True)
    # sources=rz90.apply(rx90.apply(transmitting_antenna_surface_coords))+np.array([[2.695,0,0]])
    rotation_vector1 = np.radians(np.asarray([90.0, 0.0, 0.0]))
    rotation_vector2 = np.radians(np.asarray([0.0, 0.0, -90.0]))
    transmit_horn_structure.rotate(
        o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector1),
        center=False,
    )
    transmit_horn_structure.rotate(
        o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector2),
        center=False,
    )
    transmit_horn_structure.translate(np.asarray([2.695, 0, 0]), relative=True)
    transmitting_antenna_surface_coords.rotate(
        o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector1),
        center=False,
    )
    transmitting_antenna_surface_coords.rotate(
        o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector2),
        center=False,
    )
    transmitting_antenna_surface_coords.translate(
        np.asarray([2.695, 0, 0]), relative=True
    )
    # position receiver
    receive_horn_structure.rotate(
        o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector1),
        center=False,
    )
    # rotation_vector3=np.radians(np.asarray([0.0,0.0,-90.0]))
    # receive_horn_structure.rotate(o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector3),center=False)
    receive_horn_structure.translate(np.asarray([0, 1.427, 0]), relative=True)
    recieving_antenna_surface_coords.rotate(
        o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector1),
        center=False,
    )
    # recieving_antenna_surface_coords.rotate(o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector3),center=False)
    recieving_antenna_surface_coords.translate(np.asarray([0, 1.427, 0]), relative=True)
    # sinks=rx90.apply(recieving_antenna_surface_coords)+np.array([[0.0,1.427,0]])
    source_points = transmitting_antenna_surface_coords
    sink_points = recieving_antenna_surface_coords
    receiving_antenna_points = len(np.asarray(recieving_antenna_surface_coords.points))
    transmitting_antenna_points = len(
        np.asarray(transmitting_antenna_surface_coords.points)
    )
    angle_d = angle_values[angle_inc]
    # angle_d=45
    # resolution=4
    # create environment triangle data
    # scatter_points=testtargets()
    # temp_environment=testbox()
    # reflectorplate=rectReflector(0.3,0.3,0.01)
    if resolution == 0:
        # reflectorplate=rectReflector(1.0,1.0,0.1)
        reflectorplate, scatter_points = TL.meshedReflector(
            0.3, 0.15, 6e-3, wavelength * 0.5, sides="front"
        )
        filename = "SmallTallReflector"
        position_vector = np.asarray([29e-3, 0.0, 0])
    elif resolution == 1:
        reflectorplate, scatter_points = TL.meshedReflector(
            0.15, 0.3, 6e-3, wavelength * 0.5, sides="front"
        )
        filename = "SmallWideReflector"
        position_vector = np.asarray([29e-3, 0.0, 0])
        # reflectorplate=o3d.io.read_triangle_mesh("ReferenceSquare1Meshed.stl")
    elif resolution == 2:
        # point_area=(mesh_resolution)**2
        # reflectorplate=copy.deepcopy(reflectorplate_main)
        # scatter_points=copy.deepcopy(scatter_points_main)
        reflectorplate, scatter_points = TL.meshedReflector(
            0.3, 0.3, 6e-3, wavelength * 0.5, sides="front"
        )
        # scatter_points=TL.meshsolid(reflectorplate, point_area)
        filename = "MediumSquareReflectorEMtest"
        position_vector = np.asarray([29e-3, 0.0, 0])
        # position_vector=np.asarray([0,0.0,0])
        # reflectorplate=o3d.io.read_triangle_mesh("ReferenceSquare0p5Meshed.stl")
    elif resolution == 3:
        reflectorplate, scatter_points = TL.meshedReflector(
            0.3, 0.6, 6e-3, wavelength * 0.5
        )
        filename = "LargeWideReflector"
        position_vector = np.asarray([29e-3, 0.0, 0])
        # reflectorplate=o3d.io.read_triangle_mesh("ReferenceSquare0p1Meshed.stl")
    elif resolution == 4:
        reflectorplate, scatter_points = TL.meshedReflector(
            0.6, 0.3, 6e-3, wavelength * 0.5
        )
        filename = "LargeTallReflector"
        position_vector = np.asarray([29e-3, 0.0, 0])
        # reflectorplate=o3d.io.read_triangle_mesh("ReferenceSquare0p1Meshed.stl")
        # scatter_points=o3d.geometry.PointCloud()
        # reflectorplate.compute_vertex_normals()
        # scatter_points.points=o3d.utility.Vector3dVector(np.asarray(reflectorplate.vertices)+np.asarray(reflectorplate.vertex_normals)*EPSILON)
    elif resolution == 5:
        # reflectorplate,scatter_points=RF.parabolic_reflector(0.6,0.147,6e-3,wavelength*5,sides='front')
        # reflectorplate,scatter_points=TL.parabolic_reflector(0.3,0.147,6e-3,wavelength*0.5,sides='front')
        reflectorplate, scatter_points = TL.parabola(0.6, 0.147, 6e-3, mesh_resolution)
        position_vector = np.asarray([0.0, 0.0, 17e-2])
        filename = "ParabaloidReflector"
    elif resolution == 6:
        reflectorplate, scatter_points = RD.mediumreferencescan(mesh_resolution)
        position_vector = np.asarray([0, 0, 29e-3])
        filename = "LaserScannedMediumReference"
    elif resolution == 7:
        reflectorplate, scatter_points = RD.prototypescan(mesh_resolution)
        position_vector = np.asarray([29e-3, 0.0, 0])
        filename = "LaserScannedPrototypeReflector"

    rotation_vector1 = np.radians(np.asarray([0.0, 90.0, 0.0]))
    scatter_cloud = scatter_points.rotate(
        o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector1),
        center=False,
    )
    reflectorplate.rotate(
        o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector1),
        center=False,
    )
    reflectorplate.translate(position_vector, relative=True)
    scatter_points.translate(position_vector, relative=True)

    # positionedreflector1=transformO3dobject(reflectorplate,position_vector,rotation_vector1)
    # scatter_cloud_positioned=transformO3dobject(scatter_points,position_vector,rotation_vector1)

    rotation_vector = np.radians(np.asarray([0.0, 0.0, angle_d]))
    scatter_cloud.rotate(
        o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector),
        center=False,
    )
    reflectorplate.rotate(
        o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector),
        center=False,
    )

    # testobject=coneReflector(2,2)
    reflectorplate.compute_vertex_normals()
    reflectorplate.compute_triangle_normals()
    solids = reflectorplate + transmit_horn_structure + receive_horn_structure
    temp_environment = np.append(
        np.append(
            RF.convertTriangles(reflectorplate),
            RF.convertTriangles(transmit_horn_structure),
            axis=0,
        ),
        RF.convertTriangles(receive_horn_structure),
        axis=0,
    )
    ENV_TRI_NUM = len(temp_environment)
    # sources=np.zeros((1,3),dtype=np.float32)
    # sinks=np.zeros((1,3),dtype=np.float32)

    # scatter_cloud=TL.meshsolid(reflectorplate, point_area)
    scattering_points = np.asarray(scatter_cloud.points)
    # scatter_cloud=RF.points2pointcloud(scattering_points)

    if show_scene:
        mesh_frame = o3d.geometry.TriangleMesh.create_coordinate_frame(
            size=0.5, origin=[0, 0, 0]
        )
        o3d.visualization.draw_geometries(
            [
                source_points,
                sink_points,
                scatter_cloud,
                reflectorplate,
                mesh_frame,
                receive_horn_structure,
                transmit_horn_structure,
            ]
        )
        # o3d.visualization.draw_geometries([source_points,sink_points,scatter_cloud,mesh_frame])
    ray_start = timer()
    # desired_E_axis=np.zeros((3,3),dtype=np.float32)
    # desired_E_axis[0,0]=1.0
    # desired_E_axis[1,1]=1.0
    # desired_E_axis[2,2]=1.0
    desired_E_axis = np.zeros((1, 3), dtype=np.float32)
    desired_E_axis[0, 2] = 1.0
    # responsex2[angle_inc,:],responsey2[angle_inc,:],responsez2[angle_inc,:]=ml.calculate_scattering(transmitting_antenna_surface_coords,
    #                                                                                                recieving_antenna_surface_coords,
    #                                                                                                solids,
    #                                                                                                desired_E_axis,
    #                                                                                                scatter_points=scatter_cloud,
    #                                                                                                wavelength=wavelength,
    #                                                                                                scattering=1)
    Ex, Ey, Ez = FD.calculate_scattering(
        aperture_coords=transmitting_antenna_surface_coords,
        sink_coords=recieving_antenna_surface_coords,
        antenna_solid=solids,
        desired_E_axis=desired_E_axis,
        scatter_points=scatter_cloud,
        wavelength=wavelength,
        scattering=1,
    )
    isoresponse[angle_inc] = FD.calculate_scattering_isotropic(
        aperture_coords=transmitting_antenna_surface_coords,
        sink_coords=recieving_antenna_surface_coords,
        antenna_solid=solids,
        scatter_points=scatter_cloud,
        wavelength=wavelength,
        scattering=1,
    )
    responsex2[angle_inc, :] = Ex
    responsey2[angle_inc, :] = Ey
    responsez2[angle_inc, :] = Ez

    # print((timer2-timer1)/(timer3-timer2))
    # print("{:7.0f} Rays Processed in {:3.1f}s by version 1".format(total_rays,(start-ray_start)+(timer2-timer1)) )
    # print("{:7.0f} Rays Processed in {:3.1f}s by version 2".format(total_rays,(start-ray_start)+(timer3-timer2)) )
    # print("Progress:  {:3.1f}%".format(100*(angle_inc+1)/len(angle_values)) )
    # print("Final Processing Time:  {:3.1f} s, Total time: {:3.1f} s".format(kernel_dt,total_time) )


# normalised_max=np.max(np.array([np.max(20*np.log10(np.abs(responsex3))),np.max(20*np.log10(np.abs(responsey3))),np.max(20*np.log10(np.abs(responsez3)))]))
# np.savez(filename, responsex=responsex2,responsey=responsey2,responsez=responsez2,angle_values=angle_values)
excitaiton = 2

# normalised_max2=np.max(np.array([np.max(10*np.log10(np.abs(responsex3))),np.max(10*np.log10(np.abs(responsey3))),np.max(10*np.log10(np.abs(responsez3)))]))
normalised_max3 = np.max(
    np.array(
        [
            np.max(20 * np.log10(np.abs(responsex2))),
            np.max(20 * np.log10(np.abs(responsey2))),
            np.max(20 * np.log10(np.abs(responsez2))),
        ]
    )
)
testx = 20 * np.log10(np.abs(responsex2[:])) - normalised_max3
testy = 20 * np.log10(np.abs(responsey2[:])) - normalised_max3
testz = 20 * np.log10(np.abs(responsez2[:])) - normalised_max3
# testxtd=20*np.log10(np.abs(np.sum(Ex[:,excitation,:],axis=1)))
# testytd=20*np.log10(np.abs(np.sum(Ey[:,excitation,:],axis=1)))
# testztd=20*np.log10(np.abs(np.sum(Ez[:,excitation,:],axis=1)))
testiso = 20 * np.log10(np.abs(isoresponse))
# testiso2=20*np.log10(np.abs(responseiso2))
# test1=20*np.log10(np.abs(response[:,0]))
# test2=20*np.log10(np.abs(response[:,1]))
# test3=20*np.log10(np.abs(response[:,2]))
# test2=20*np.log10(np.abs(response[:,1]))-np.max(np.max(20*np.log10(np.abs(response[:,1]))))
# test3=20*np.log10(np.abs(response[:,2]))-np.max(np.max(20*np.log10(np.abs(response[:,2]))))
# test4=20*np.log10(np.abs(response[:,3]))-np.max(np.max(20*np.log10(np.abs(response[:,3]))))
# test5=20*np.log10(np.abs(response[:,4]))-np.max(np.max(20*np.log10(np.abs(response[:,4]))))
# test5=20*np.log10(np.abs(response[:,0])-np.max(np.max(np.abs(response[:,4]))))
# test4=20*np.log10(np.abs(response[:,3]))
# test5=20*np.log10(np.abs(response[:,4]))
# test2=20*np.log10(np.abs(response[:,1]))-np.max(np.max(20*np.log10(np.abs(response[:,1]))))
# test5=20*np.log10(np.abs(response[:,4]))-np.max(np.max(20*np.log10(np.abs(response[:,4]))))
# plt.plot(angle_values-45.0,test4,angle_values-45.0,test5)
# Create plots with pre-defined labels.
# fig, ax = plt.subplots()
# ax.plot(angle_values-45, 20*np.log10(np.abs(responsez2[:,1]))-np.max(20*np.log10(np.abs(responsez2[:,1]))),label='Ez')
# ax.plot(angle_values-45, 20*np.log10(np.abs(responsez[:,1]))-np.max(20*np.log10(np.abs(responsez[:,1]))),label='Ez')
# ax.plot(angle_values-45, testx[:,2],label='Ex')
# ax.plot(angle_values-45, testy[:,2],label='Ey')
# ax.plot(angle_values-46, testz[:,2],label='FD')
# ax.plot(angle_values-45, testx,label='Extd')
# ax.plot(angle_values-45, testy,label='Eytd')
# ax.plot(angle_values-45, testz,label='Eztd')
# ax.plot(angle_values-45, testiso-np.max(testiso), linestyle=':', label='No Polarization')
# ax.plot(angle_values-45, testiso2-np.max(testiso2), linestyle=':', label='No Polariztion')
# ax.plot(angle_values-45.0, test3, label='$Medium$')
# ax.plot(angle_values-45.0, test4, label='$Large T$')
# ax.plot(angle_values-45.0, test5, label='$Large W$')
# plt.xlabel('$\\theta_{N}$ (degrees)')
# plt.ylabel('Normalised Level (dB)')
# plt.title('Scattering=fraction')
# ax.set_ylim(-60.0,0)
# ax.set_xlim(np.min(angle_values)-45,np.max(angle_values)-45)
# ax.set_xticks(np.linspace(np.min(angle_values)-45, np.max(angle_values)-45, 19))
# ax.set_xticks(np.linspace(0, 360, 19))
# ax.set_yticks(np.linspace(-60, 0., 21))
# legend = ax.legend(loc='upper right', shadow=True)
# plt.grid()
# plt.show()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jun 16 13:15:07 2020

@author: timtitan
"""
import logging
from timeit import default_timer as timer

import matplotlib.pyplot as plt
import numpy as np
import open3d as o3d
import scipy.signal as sig
# import GPUScatteringTest as GU
import scipy.stats
from scipy.fft import fft, fftfreq
from tqdm import tqdm

import lyceanem.geometry.targets as TL
import lyceanem.models.time_domain as TD
import lyceanem.raycasting.rayfunctions as RF
import lyceanem.tests.reflectordata as reflectordata

nb_logger = logging.getLogger('numba')
nb_logger.setLevel(logging.ERROR)  # only show error
#angle_values=np.linspace(0,90,181)
max_scatter=2
sampling_freq=60e9
model_time=1e-7
num_samples=int(model_time*(sampling_freq))
#time_record=np.zeros((len(angle_values),num_samples,3),dtype='float')
#simulate receiver noise
bandwidth=8e9
kb=1.38065e-23
receiver_impedence=50
thermal_noise_power=4*kb*293.15*receiver_impedence*bandwidth
noise_power=-80#dbw
mean_noise=0

#noise_volts = np.random.normal(mean_noise, np.sqrt(thermal_noise_power*receiver_impedence), num_samples)
#time_ref=np.zeros((len(angle_values)),dtype='float')
model_freq=26e9
wavelength=3e8/model_freq
show_scene=True
transmit_horn_structure,transmitting_antenna_surface_coords=TL.meshedHorn(58e-3,58e-3,128e-3,2e-3,0.21,0.5*wavelength)
rotation_vector1=np.radians(np.asarray([90.0,0.0,0.0]))
rotation_vector2=np.radians(np.asarray([0.0,0.0,-90.0]))
transmit_horn_structure.rotate(o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector1),center=False)
transmit_horn_structure.rotate(o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector2),center=False)
transmitting_antenna_surface_coords.rotate(o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector1),center=False)
transmitting_antenna_surface_coords.rotate(o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector2),center=False)
transmit_horn_structure.translate(np.asarray([2.695,0,0]),relative=True)
transmitting_antenna_surface_coords.translate(np.asarray([2.695,0,0]),relative=True)
source_coords=np.asarray(transmitting_antenna_surface_coords.points)
source_normals=np.asarray(transmitting_antenna_surface_coords.normals)
receive_horn_structure,recieving_antenna_surface_coords=TL.meshedHorn(58e-3,58e-3,128e-3,2e-3,0.21,0.5*wavelength)
receive_horn_structure.rotate(o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector1),center=False)
#rotation_vector3=np.radians(np.asarray([0.0,0.0,-90.0]))
#receive_horn_structure.rotate(o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector3),center=False)
receive_horn_structure.translate(np.asarray([0,1.427,0]),relative=True)
recieving_antenna_surface_coords.rotate(o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector1),center=False)
#recieving_antenna_surface_coords.rotate(o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector3),center=False)
recieving_antenna_surface_coords.translate(np.asarray([0,1.427,0]),relative=True)

#source_normals[0,2]=1.0
#sinks=np.zeros((1,3),dtype=np.float32)    
#sinks[0,2]=1.0
#sink_normals=np.zeros((1,3),dtype=np.float32)  
#sink_normals[0,2]=-1.0
resolution=2
angles=np.linspace(0,90,91)
wake_times=np.zeros((len(angles)))
Ex=np.zeros((len(angles),num_samples))
Ey=np.zeros((len(angles),num_samples))
Ez=np.zeros((len(angles),num_samples))
time_points=np.zeros((len(angles)))
for angle in tqdm(range(len(angles))):
    start_angle=timer()
    rotation_vector3=np.radians(np.asarray([0.0,90.0,0.0]))
    rotation_vector4=np.radians(np.asarray([0.0,0.0,angles[angle]]))
    #reflectorplate,scatter_points=TL.meshedReflectorv2(0.3,0.3,6e-3,0.5*wavelength,sides='all')
    if(resolution==0):
        #reflectorplate=rectReflector(1.0,1.0,0.1)
        reflectorplate,scatter_points=TL.meshedReflector(0.3,0.15,6e-3,wavelength*0.5,sides='front')
        filename='SmallTallReflector'
        position_vector=np.asarray([29e-3,0.0,0])
    elif(resolution==1):
        reflectorplate,scatter_points=TL.meshedReflector(0.15,0.3,6e-3,wavelength*0.5,sides='front')
        filename='SmallWideReflector'
        position_vector=np.asarray([29e-3,0.0,0])
        #reflectorplate=o3d.io.read_triangle_mesh("ReferenceSquare1Meshed.stl")            
    elif(resolution==2):
        #point_area=(mesh_resolution)**2
        #reflectorplate=copy.deepcopy(reflectorplate_main)
        #scatter_points=copy.deepcopy(scatter_points_main)
        reflectorplate,scatter_points=TL.meshedReflector(0.3,0.3,6e-3,wavelength*0.5,sides='all')
        #scatter_points=TL.meshsolid(reflectorplate, point_area)
        filename='MediumSquareReflectorEMtest'
        position_vector=np.asarray([29e-3,0.0,0])
        #position_vector=np.asarray([0,0.0,0])
        #reflectorplate=o3d.io.read_triangle_mesh("ReferenceSquare0p5Meshed.stl")
    elif(resolution==3):
        reflectorplate,scatter_points=TL.meshedReflector(0.3,0.6,6e-3,wavelength*0.5)
        filename='LargeWideReflector'
        position_vector=np.asarray([29e-3,0.0,0])
        #reflectorplate=o3d.io.read_triangle_mesh("ReferenceSquare0p1Meshed.stl")
    elif(resolution==4):
        reflectorplate,scatter_points=TL.meshedReflector(0.6,0.3,6e-3,wavelength*0.5)
        filename='LargeTallReflector'
        position_vector=np.asarray([29e-3,0.0,0])
        #reflectorplate=o3d.io.read_triangle_mesh("ReferenceSquare0p1Meshed.stl")
        #scatter_points=o3d.geometry.PointCloud()
        #reflectorplate.compute_vertex_normals()
        #scatter_points.points=o3d.utility.Vector3dVector(np.asarray(reflectorplate.vertices)+np.asarray(reflectorplate.vertex_normals)*EPSILON)
    elif(resolution==5):
        #reflectorplate,scatter_points=RF.parabolic_reflector(0.6,0.147,6e-3,wavelength*5,sides='front')
        #reflectorplate,scatter_points=TL.parabolic_reflector(0.3,0.147,6e-3,wavelength*0.5,sides='front')
        reflectorplate,scatter_points=TL.parabola(0.6,0.147,6e-3,wavelength*0.5)
        position_vector=np.asarray([0.0,0.0,17e-2])
        filename='ParabaloidReflector'
    elif(resolution==6):
        reflectorplate,scatter_points=reflectordata.mediumreferencescan(wavelength*0.5)
        position_vector=np.asarray([0,0,29e-3])
        filename='LaserScannedMediumReference'
    elif(resolution==7):
        reflectorplate,scatter_points=reflectordata.prototypescan(wavelength*0.5)
        position_vector=np.asarray([29e-3,0.0,0])
        filename='LaserScannedPrototypeReflector'
    #reflectorplate,scatter_points=TL.meshedReflector(0.3,0.3,6e-3,wavelength*0.5,sides='front')
    #filename='SmallTallReflector'
    position_vector=np.asarray([29e-3,0.0,0])
    #reflectorplate.subdivide_midpoint(4)
    reflectorplate.compute_triangle_normals()
    #position_vector=np.asarray([29e-3,0.0,0])
    #position_vector=np.asarray([0,0.0,0])
    reflectorplate.rotate(o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector3),center=False)
    reflectorplate.translate(position_vector)
    reflectorplate.rotate(o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector4),center=False)
    scatter_points.rotate(o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector3),center=False)
    scatter_points.translate(position_vector)
    scatter_points.rotate(o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector4),center=False)
    #reflectorplate.translate([-angles[angle]/100,0,0])
    #scatter_points.translate([-angles[angle]/100,0,0])
    noise_volts = np.random.normal(mean_noise, np.sqrt(thermal_noise_power*receiver_impedence), num_samples)
    scattering_points=np.asarray(scatter_points.points)
    scattering_normals=np.asarray(scatter_points.normals)
    mesh_frame = o3d.geometry.TriangleMesh.create_coordinate_frame(size=0.5, origin=[0, 0, 0])
    temp_environment=RF.convertTriangles(reflectorplate)
    #full_index=RF.workchunkingv2(source_coords,sinks,scattering_points,temp_environment,max_scatter)
    #o3d.visualization.draw_geometries([mesh_frame,recieving_antenna_surface_coords,transmitting_antenna_surface_coords,reflectorplate,scatter_points,transmit_horn_structure])
    #for excitation_index in range(3):
    excitation_index=1
    #process for scattering matrix
    #angle_response[:,:,incrementer,plate,angle_inc],impulse_response[:,:,incrementer,plate,angle_inc],max_time[incrementer,plate,angle_inc]=RF.CalculatePoyntingVectors(total_network,wavelength,full_index,az_bins=np.linspace(-np.pi,np.pi,361),impulse=False,aoa=True)
    #scatter_map_test=RF.VectorNetworkProcessv2(np.asarray(transmitting_antenna_surface_coords.points),np.asarray(recieving_antenna_surface_coords.points),scattering_points,full_index,wavelength)
        #scatter_map=RF.VectorNetworkProcessEM(source_coords.shape[0],sinks.shape[0],unified_model,unified_normals,unified_weights,point_informationv2,full_index,scattering_coefficient.astype(np.complex64),wavelength)
    #scatter_map=RF.VectorNetworkProcessTimeEM(source_coords.shape[0],sinks.shape[0],unified_model,unified_normals,unified_weights,point_informationv2,full_index,wavelength)
    #responsex2[angle_inc,excitation_index]=np.dot(np.dot(np.ones((transmitting_antenna_points)),scatter_map[:,:,0,1]),np.ones((receiving_antenna_points)))
    #responsey2[angle_inc,excitation_index]=np.dot(np.dot(np.ones((transmitting_antenna_points)),scatter_map[:,:,1,1]),np.ones((receiving_antenna_points)))
    #responsez2[angle_inc,excitation_index]=np.dot(np.dot(np.ones((transmitting_antenna_points)),scatter_map[:,:,2,1]),np.ones((receiving_antenna_points)))
    #responsex[angle_inc,excitation_index]=np.dot(np.dot(np.ones((transmitting_antenna_points)),np.sum(scatter_map[:,:,0,:],axis=2)),np.ones((receiving_antenna_points)))
    #responsey[angle_inc,excitation_index]=np.dot(np.dot(np.ones((transmitting_antenna_points)),np.sum(scatter_map[:,:,1,:],axis=2)),np.ones((receiving_antenna_points)))
    #responsez[angle_inc,excitation_index]=np.dot(np.dot(np.ones((transmitting_antenna_points)),np.sum(scatter_map[:,:,2,:],axis=2)),np.ones((receiving_antenna_points)))
    #paths=EM.EMGPUPathLengths(source_coords.shape[0],sinks.shape[0],full_index,point_informationv2)
    #polar_coefficients=EM.EMGPUPolarMixing(source_coords.shape[0],sinks.shape[0],full_index,point_informationv2)
    #ray_components=EM.EMGPUWrapper(num_sources,num_sinks,full_index,point_informationv2,wavelength)
    #loss=EM.pathloss(paths,wavelength)
    #full_rays=loss.reshape(loss.shape[0],1)*polar_coefficients
    #depth_slicelos=full_index[np.equal(full_index[:,2],0),:][:,[0,1]]
    #depth_slicebounce=full_index[~np.equal(full_index[:,2],0),:][:,[0,2]]
    #scatter_iso_map=RF.scatter_net_sorttest(source_coords.shape[0],sinks.shape[0],np.zeros((source_coords.shape[0],sinks.shape[0],2),dtype=np.complex64),depth_slicelos,loss[np.equal(full_index[:,2],0)],0)
    #scatter_iso_map=RF.scatter_net_sorttest(source_coords.shape[0],sinks.shape[0],scatter_iso_map,depth_slicebounce,loss[~np.equal(full_index[:,2],0)],1)
    #responseiso2[angle_inc]=np.dot(np.dot(np.ones((source_coords.shape[0])),np.sum(scatter_iso_map[:,:,:],axis=2)),np.ones((sinks.shape[0])))
    #scatter_map2=RF.scatter_net_sortEM(source_coords.shape[0],sinks.shape[0],np.zeros((source_coords.shape[0],sinks.shape[0],3,2),dtype=np.complex64),depth_slicelos,full_rays[np.equal(full_index[:,2],0),:],0)
    #scatter_map2=RF.scatter_net_sortEM(source_coords.shape[0],sinks.shape[0],scatter_map2,depth_slicebounce,full_rays[~np.equal(full_index[:,2],0),:],1)
    #time_index=np.linspace(0,model_time,num_samples)
    pulse_time=5e-9
    output_power=0.01 #dBwatts
    powerdbm=10*np.log10(output_power)+30
    v_transmit=((10**(powerdbm/20))*receiver_impedence)**0.5
    output_amplitude_rms=v_transmit/(1/np.sqrt(2))
    output_amplitude_peak=v_transmit
    #excitation_signal=sig.unit_impulse(100)
   
    desired_E_axis=np.zeros((3),dtype=np.float32)
    desired_E_axis[2]=1.0
    noise_volts_peak=(10**(noise_power/10)*receiver_impedence)*0.5
    #o3d.visualization.draw_geometries([mesh_frame,reflectorplate,scatter_hit_points,transmit_horn_structure])
    #print(angles[angle],len(full_index[full_index[:,2]!=0,:]))
    excitation_signal=output_amplitude_rms*sig.chirp(np.linspace(0,
                                                                 pulse_time,
                                                                 int(pulse_time*sampling_freq)),
                                                                 model_freq-bandwidth, 
                                                                 pulse_time, 
                                                                 model_freq, 
                                                                 method='linear', 
                                                                 phi=0, 
                                                                 vertex_zero=True)+np.random.normal(mean_noise, noise_volts_peak, int(pulse_time*sampling_freq))
    Ex[angle,:],Ey[angle,:],Ez[angle,:],wake_times[angle]=TD.calculate_scattering(transmitting_antenna_surface_coords,
                                                                recieving_antenna_surface_coords,
                                                                excitation_signal,
                                                                reflectorplate,
                                                                desired_E_axis,
                                                                scatter_points=scatter_points,
                                                                wavelength=wavelength,
                                                                scattering=1,
                                                                elements=False,
                                                                sampling_freq=sampling_freq,
                                                                num_samples=num_samples)
    
    noise_volts=np.random.normal(mean_noise, noise_volts_peak, num_samples)
    Ex[angle,:]=Ex[angle,:]+noise_volts
    Ey[angle,:]=Ey[angle,:]+noise_volts
    Ez[angle,:]=Ez[angle,:]+noise_volts
    time_points[angle]=timer()-start_angle
    #est_remaining=np.mean(time_points[:(angle+1)])*(len(angles)-(angle+1))
    #print("Processing Time:  {:3.1f} s, Estimated Remaining time: {:3.1f} minutes".format(time_points[angle],est_remaining/60) )
    #np.savez('timedomainsmallplatechirp',Ex=Ex,Ey=Ey,Ez=Ez,excitation_signal=excitation_signal,time_index=time_index)
    
norm_max=np.nanmax(np.array([np.nanmax(10*np.log10((Ex**2)/receiver_impedence)),np.nanmax(10*np.log10((Ey**2)/receiver_impedence)),np.nanmax(10*np.log10((Ez**2)/receiver_impedence))]))
np.savez(filename,Ex=Ex,Ey=Ey,Ez=Ez,excitation_signal=excitation_signal,time_index=time_index)
#ax.plot(time_index,time_map[0,0,:])
#ax.plot(time_index,20*np.log10(Ex)-norm_max)
#ax.plot(time_index,20*np.log10(Ey)-norm_max)
#ax.plot(time_index,20*np.log10(Ez[0,:])-norm_max)
#ax.plot(time_index,20*np.log10(Ez[1,:]**2)-norm_max)
#ax.plot(time_index,20*np.log10(Ez[2,:]**2)-norm_max)
#ax.plot(time_index,20*np.log10(Ez[91,:]**2)-norm_max)
fig,ax=plt.subplots()
ax.plot(angles,10*np.log10(np.sum((Ex**2/receiver_impedence),axis=1)))
ax.plot(angles,10*np.log10(np.sum((Ey**2/receiver_impedence),axis=1)))
ax.plot(angles,10*np.log10(np.sum((Ez**2/receiver_impedence),axis=1)))
time_index=np.linspace(0,model_time*1e9,num_samples)
time,anglegrid=np.meshgrid(time_index[:1801],angles-45)
#fig2 = plt.figure()
#ax2 = plt.axes(projection='3d')
fig2, ax2 = plt.subplots(constrained_layout=True)
origin = 'lower'
# Now make a contour plot with the levels specified,
# and with the colormap generated automatically from a list
# of colors.

levels = np.linspace(-80,0,41)
#CS = ax2.plot_surface(anglegrid,time, 10*np.log10((Ez[:,:1801]**2)/receiver_impedence)-norm_max,levels,
#                    origin=origin,
#                    extend='both')
CS = ax2.contourf(anglegrid,time, 10*np.log10((Ez[:,:1801]**2)/receiver_impedence)-norm_max, levels,
                  origin=origin,
                  extend='both')
cbar = fig2.colorbar(CS)
cbar.ax.set_ylabel('Received Power (dBm)')
#cbar.set_clim(-60,0)
#cbar.set_ticks([norm_max-60,norm_max])
#cbar.ax.set_yticklabels([norm_max-60,norm_max])
ax2.set_ylim(0,30)
ax2.set_xlim(-45,45)
#ax2.set_zlim(-60,0)
ax2.set_xticks(np.linspace(-45, 45, 7))
ax2.set_yticks(np.linspace(0, 30, 16))
#ax2.set_yticklabels(np.linspace(0, np.max(time_index)/1e-9, 11).astype(str))
ax2.set_xlabel('Rotation Angle (degrees)')
ax2.set_ylabel('Time of Flight (ns)')
ax2.set_title('Received Power vs Time for rotating Plate (24GHz)')
#ax2.grid()
#ax.plot(time_index,time_map[0,2,:])
#graphname=filename+'Graph.pdf'
#fig.savefig(graphname)   # save the figure to file
#plt.close(fig)    # close the figure window
# testx2=20*np.log10(np.abs(responsex2[:,2]))
# testy2=20*np.log10(np.abs(responsey2[:,2]))
# testz2=20*np.log10(np.abs(responsez2[:,2]))
# testiso=20*np.log10(np.abs(responseiso))
# #test1=20*np.log10(np.abs(response[:,0]))
# #test2=20*np.log10(np.abs(response[:,1]))
# #test3=20*np.log10(np.abs(response[:,2]))
# #test2=20*np.log10(np.abs(response[:,1]))-np.max(np.max(20*np.log10(np.abs(response[:,1]))))
# #test3=20*np.log10(np.abs(response[:,2]))-np.max(np.max(20*np.log10(np.abs(response[:,2]))))
# #test4=20*np.log10(np.abs(response[:,3]))-np.max(np.max(20*np.log10(np.abs(response[:,3]))))
# #test5=20*np.log10(np.abs(response[:,4]))-np.max(np.max(20*np.log10(np.abs(response[:,4]))))
# #test5=20*np.log10(np.abs(response[:,0])-np.max(np.max(np.abs(response[:,4]))))
# #test4=20*np.log10(np.abs(response[:,3]))
# #test5=20*np.log10(np.abs(response[:,4]))
# #test2=20*np.log10(np.abs(response[:,1]))-np.max(np.max(20*np.log10(np.abs(response[:,1]))))
# #test5=20*np.log10(np.abs(response[:,4]))-np.max(np.max(20*np.log10(np.abs(response[:,4]))))
# #plt.plot(angle_values-45.0,test4,angle_values-45.0,test5)
# # Create plots with pre-defined labels.
# normalised_max=np.max(np.array([np.max(20*np.log10(np.abs(responsex2[:,0]))),np.max(20*np.log10(np.abs(responsex2[:,1]))),np.max(20*np.log10(np.abs(responsex2[:,2]))),np.max(20*np.log10(np.abs(responsex2[:,2]))),np.max(20*np.log10(np.abs(responseiso)))]))
#fig, ax = plt.subplots()
#ax.plot(time_record[0,time_arrive:time_arrive+3000,0],label='Ex')
#ax.plot(time_record[0,time_arrive:time_arrive+3000,1],label='Ey')
#ax.plot(time_record[0,time_arrive:time_arrive+3000,2],label='Ez')
# ax.plot(angle_values-45, testy2-normalised_max,label='Ey')
# ax.plot(angle_values-45, testz2-normalised_max,label='Ez')
#fig, ax = plt.subplots()
#ax.plot(time_record[45,time_arrive:time_arrive+3000,0],label='Ex')
#ax.plot(time_record[45,time_arrive:time_arrive+3000,1],label='Ey')
#ax.plot(time_record[45,time_arrive:time_arrive+3000,2],label='Ez2')
# ax.plot(angle_values-45, testiso-normalised_max, linestyle=':', label='No Polariztion')
# #ax.plot(angle_values-45.0, test3, label='$Medium$')
# #ax.plot(angle_values-45.0, test4, label='$Large T$')
# #ax.plot(angle_values-45.0, test5, label='$Large W$')
# plt.xlabel('$\\theta_{N}$ (degrees)')
# plt.ylabel('Normalised Level (dB)')
# plt.title('Ex Excitation')
# ax.set_ylim(-100.0,0)
# ax.set_xlim(-45.0,45)
# ax.set_xticks(np.linspace(-45, 90-45, 19))
# ax.set_yticks(np.linspace(-100, 0., 21))
# legend = ax.legend(loc='upper right', shadow=True)
# plt.grid()
# plt.show()
freq,psd=sig.welch(Ez,fs=sampling_freq)
freqgrid,anglegrid=np.meshgrid(freq,angles-45)
fig2, ax2 = plt.subplots(constrained_layout=True)
origin = 'lower'
# Now make a contour plot with the levels specified,
# and with the colormap generated automatically from a list
# of colors.

levels = np.linspace(0,1,61)
CS = ax2.contourf(anglegrid,freqgrid/1e9, psd/np.max(psd), 
                  levels, 
                  origin=origin,
                  extend='both')
cbar = fig2.colorbar(CS)
cbar.ax.set_ylabel('Power Spectral Density')
#cbar.set_ticks([norm_max-60,norm_max])
#cbar.ax.set_yticklabels([norm_max-60,norm_max])
ax2.set_ylim(2,30)
ax2.set_xlim(-45,45)
ax2.set_xticks(np.linspace(-45, 45, 7))
ax2.set_yticks(np.linspace(2, 30, 15))
#ax2.set_yticklabels(np.linspace(0, np.max(time_index)/1e-9, 11).astype(str))
ax2.set_xlabel('Rotation Angle (degrees)')
ax2.set_ylabel('Frequency (Hz)')
ax2.set_title('Power Spectral Density for rotating Plate (24GHz)')
xf = fftfreq(len(time_index), 1/sampling_freq)[:len(time_index)//2]
yf=fft(Ez)
w = sig.blackman(6000)
ywf = fft(Ez*w)
freqgrid,anglegrid=np.meshgrid(xf[1:len(time_index)//2],angles-45)
fig2, ax2 = plt.subplots(constrained_layout=True)
origin = 'lower'
# Now make a contour plot with the levels specified,
# and with the colormap generated automatically from a list
# of colors.

levels = np.linspace(-60,0,61)
CS = ax2.contourf(anglegrid,freqgrid/1e9, 10*np.log10(np.abs(ywf[:,1:len(time_index)//2]))-np.nanmax(10*np.log10(np.abs(ywf[:,1:len(time_index)//2]))), 
                  levels, 
                  origin=origin,
                  extend='both')
cbar = fig2.colorbar(CS)
cbar.ax.set_ylabel('Power (dBm)')
#cbar.set_ticks([norm_max-60,norm_max])
#cbar.ax.set_yticklabels([norm_max-60,norm_max])
ax2.set_ylim(2,30)
ax2.set_xlim(-45,45)
ax2.set_xticks(np.linspace(-45, 45, 7))
ax2.set_yticks(np.linspace(2, 30, 15))
#ax2.set_yticklabels(np.linspace(0, np.max(time_index)/1e-9, 11).astype(str))
ax2.set_xlabel('Rotation Angle (degrees)')
ax2.set_ylabel('Frequency (GHz)')
ax2.set_title('Power Spectral Density for rotating Plate (24GHz)')

#setfreq_psd=20*np.log10(np.abs(ywf[:,1500]))
setfreq_psdalt=20*np.log10(np.abs(yf[:,1500]))
fig,ax=plt.subplots()
#ax.plot(angles,setfreq_psd)
ax.plot(angles,setfreq_psdalt)

input_signal=excitation_signal*(output_amplitude_peak)
inputfft=fft(input_signal)
input_freq=fftfreq(120,1/sampling_freq)[:60]
freqfuncabs=scipy.interpolate.interp1d(input_freq,np.abs(inputfft[:60]))
freqfuncangle=scipy.interpolate.interp1d(input_freq,np.angle(inputfft[:60]))
newinput=freqfuncabs(xf[:2950])*np.exp(freqfuncangle(xf[:2950]))
#calculate s21 at 24GHz over angles
s21=20*np.log10(np.abs(yf[:,:2950]/newinput))
tdangles=np.linspace(-45,45,91)
fig,ax=plt.subplots()
ax.plot(tdangles,s21[:,2400]-np.max(s21[:,2400]),label='24GHz')
freqcroppeds21=s21[:,2200:2601]
freqgrid,anglegrid=np.meshgrid(xf[2200:2601],angles-45)
fig2, ax2 = plt.subplots(constrained_layout=True)
origin = 'lower'
# Now make a contour plot with the levels specified,
# and with the colormap generated automatically from a list
# of colors.

levels = np.linspace(np.nanmax(20*np.log10(np.abs(freqcroppeds21[np.isfinite(freqcroppeds21)])))-10,np.nanmax(20*np.log10(np.abs(freqcroppeds21[np.isfinite(freqcroppeds21)]))),61)
CS = ax2.contourf(anglegrid,freqgrid/1e9, 20*np.log10(np.abs(freqcroppeds21)), 
                  levels, 
                  origin=origin,
                  extend='both')
cbar = fig2.colorbar(CS)
cbar.ax.set_ylabel('$S_{2,1}$ (dB)')
#cbar.set_ticks([norm_max-60,norm_max])
#cbar.ax.set_yticklabels([norm_max-60,norm_max])
ax2.set_ylim(18,30)
ax2.set_xlim(-45,45)
ax2.set_xticks(np.linspace(-45, 45, 10))
ax2.set_yticks(np.linspace(18, 30, 10))
#ax2.set_yticklabels(np.linspace(0, np.max(time_index)/1e-9, 11).astype(str))
ax2.set_xlabel('Rotation Angle (degrees)')
ax2.set_ylabel('Frequency (GHz)')
ax2.set_title('$S_{2,1}$ for rotating Plate (24GHz sine wave)')


# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import sys
import os
import re
from glob import glob
import shutil
from sphinx_gallery.scrapers import figure_rst
try:
    from importlib import metadata
except ImportError: # for Python<3.8
    import importlib_metadata as metadata
#__version__ = metadata.version("jsonschema")
#from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = metadata.version("lyceanem")
except metadata.PackageNotFoundError:
    # package is not installed
    pass
# for example take major/minor
version = '.'.join(__version__.split('.')[:2])

class PNGScraper(object):
    def __init__(self):
        self.seen = set()

    def __repr__(self):
        return 'PNGScraper'

    def __call__(self, block, block_vars, gallery_conf):
        # Find all PNG files in the directory of this example.
        path_current_example = os.path.dirname(block_vars['src_file'])
        pngs = sorted(glob(os.path.join(path_current_example, '*.png')))

        # Iterate through PNGs, copy them to the sphinx-gallery output directory
        image_names = list()
        image_path_iterator = block_vars['image_path_iterator']
        for png in pngs:
            if png not in self.seen:
                self.seen |= set(png)
                this_image_path = image_path_iterator.next()
                image_names.append(this_image_path)
                shutil.move(png, this_image_path)
        # Use the `figure_rst` helper function to generate rST for image files
        return figure_rst(image_names, gallery_conf['src_dir'])

sys.path.insert(0, os.path.abspath('.'))
sys.path.insert(0, os.path.abspath('../../'))
# -- Project information -----------------------------------------------------

project = 'LyceanEM'
copyright = '2022, Timothy Pelham'
author = 'Timothy Pelham'

# The full version, including alpha/beta/rc tags
#release = version#'0.01'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['sphinx.ext.autodoc',
              'sphinx.ext.napoleon',
              'sphinxcontrib.bibtex',
              'sphinx.ext.intersphinx',
              'sphinx.ext.viewcode',
              'sphinx_gallery.gen_gallery',
              'sphinx.ext.imgmath']
bibtex_bibfiles = ['_static/lyceanemrefs.bib']
# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

#autodoc options
#autodoc_mock_imports = ['numba',
#                        'cupy',
#                        'open3d',
#                        'solidpython']

intersphinx_mapping = {
    'python': ('https://docs.python.org/{.major}'.format(sys.version_info), None),
    'matplotlib': ('https://matplotlib.org/', None),
    'numpy': ('https://docs.scipy.org/doc/numpy', None),
    'scipy': ('https://docs.scipy.org/doc/scipy/reference', None),
    'open3d': ('http://www.open3d.org/docs/0.9.0/', None),
    'numba': ('https://numba.readthedocs.io/en/stable/', None),
    'solidpython': ('https://solidpython.readthedocs.io/en/latest/', None)
}

#Sphinx Gallery Configuration
sphinx_gallery_conf = {
    'examples_dirs':['../examples'],
    'gallery_dirs': ['auto_examples'],
    'filename_pattern': re.escape(os.sep),
    'image_scrapers': ('matplotlib', PNGScraper()),
    'matplotlib_animations': True,
    'run_stale_examples': False,
    'first_notebook_cell': ("# This cell is added by sphinx-gallery\n"
                            "# It can be customized to whatever you like\n"
                            "%matplotlib inline"),
    'last_notebook_cell': "# This is the last cell",
    'notebook_images': f'https://stonesoup.rtfd.io/en/{os.environ.get("READTHEDOCS_VERSION", "latest")}/',
    'reference_url': {
         # The module you locally document uses None
        'sphinx_gallery': None,
    },
    'plot_gallery': False, #documentation examples require cuda on build machine, so much be fully built before being passed to readthedocs
}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
#html_theme = 'classic'
html_theme = 'sphinx_rtd_theme'
html_logo = "_static/lynx.png"
html_theme_options = {
    'logo_only' : True,
    'display_version' : False,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']
source_suffix='.rst'
master_doc='index'

# -- Options for LaTeX output ---------------------------------------------
latex_engine = 'xelatex'
latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    'papersize': 'a4paper',

    # The font size ('10pt', '11pt' or '12pt').
    'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    # 'preamble': '',

    # Latex figure (float) alignment
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'LyceanEM.tex', 'LyceanEM Documentation',
     'Timothy Pelham', 'manual'),
]



import numpy as np
from numba import float32, from_dtype, njit, guvectorize


# A numpy record array (like a struct) to record triangle
point_data = np.dtype(
    [
        # conductivity, permittivity and permiability
        # free space values should be
        # permittivity of free space 8.8541878176e-12F/m
        # permeability of free space 1.25663706212e-6H/m
        ("permittivity", "c8"),
        ("permeability", "c8"),
        # electric or magnetic current sources? E if True
        ("Electric", "?"),
    ],
    align=True,
)
point_t = from_dtype(point_data)  # Create a type that numba can recognize!

# A numpy record array (like a struct) to record triangle
triangle_data = np.dtype(
    [
        # v0 data
        ("v0x", "f4"),
        ("v0y", "f4"),
        ("v0z", "f4"),
        # v1 data
        ("v1x", "f4"),
        ("v1y", "f4"),
        ("v1z", "f4"),
        # v2 data
        ("v2x", "f4"),
        ("v2y", "f4"),
        ("v2z", "f4"),
        # normal vector
        # ('normx', 'f4'),  ('normy', 'f4'), ('normz', 'f4'),
        # ('reflection', np.float64),
        # ('diffuse_c', np.float64),
        # ('specular_c', np.float64),
    ],
    align=True,
)
triangle_t = from_dtype(triangle_data)  # Create a type that numba can recognize!

# ray class, to hold the ray origin, direction, and eventuall other data.
ray_data = np.dtype(
    [
        # origin data
        ("ox", "f4"),
        ("oy", "f4"),
        ("oz", "f4"),
        # direction vector
        ("dx", "f4"),
        ("dy", "f4"),
        ("dz", "f4"),
        # target
        # direction vector
        # ('tx','f4'),('ty','f4'),('tz','f4'),
        # distance traveled
        ("dist", "f4"),
        # intersection
        ("intersect", "?"),
    ],
    align=True,
)
ray_t = from_dtype(ray_data)  # Create a type that numba can recognize!
# We can use that type in our device functions and later the kernel!

scattering_point = np.dtype(
    [
        # position data
        ("px", "f4"),
        ("py", "f4"),
        ("pz", "f4"),
        # velocity
        ("vx", "f4"),
        ("vy", "f4"),
        ("vz", "f4"),
        # normal
        ("nx", "f4"),
        ("ny", "f4"),
        ("nz", "f4"),
        # weights
        ("ex", "c8"),
        ("ey", "c8"),
        ("ez", "c8"),
        # conductivity, permittivity and permiability
        # free space values should be
        # permittivity of free space 8.8541878176e-12F/m
        # permeability of free space 1.25663706212e-6H/m
        ("permittivity", "c8"),
        ("permeability", "c8"),
        # electric or magnetic current sources? E if True
        ("Electric", "?"),
    ],
    align=True,
)

scattering_t = from_dtype(scattering_point)  # Create a type that numba can recognize!



#!/usr/bin/env python
# coding: utf-8
"""
Array Beamforming
======================================================

This example uses the frequency domain :func:`lyceanem.models.frequency_domain.calculate_farfield` function to predict
the farfield patterns for a linearly polarised aperture with multiple elements. This is then beamformed to all farfield points using multiple open loop beamforming algorithms to attemp to 'map' out the acheivable beamforming for the antenna array using :func:`lyceanem.electromagnetics.beamforming.MaximumDirectivityMap`.

The Steering Efficiency can then be evaluated using :func:`lyceanem.electromagnetics.beamforming.Steering_Efficiency` for the resultant achieved beamforming.


"""
import numpy as np
import open3d as o3d
import copy

# %%
# Setting Farfield Resolution and Wavelength
# -------------------------------------------
# LyceanEM uses Elevation and Azimuth to record spherical coordinates, ranging from -180 to 180 degrees in azimuth,
# and from -90 to 90 degrees in elevation. In order to launch the aperture projection function, the resolution in
# both azimuth and elevation is requried.
# In order to ensure a fast example, 37 points have been used here for both, giving a total of 1369 farfield points.
#
# The wavelength of interest is also an important variable for antenna array analysis, so we set it now for 10GHz,
# an X band aperture.

az_res = 181
elev_res = 37
wavelength = 3e8 / 10e9

# %%
# Geometries
# ------------------------
# In order to make things easy to start, an example geometry has been included within LyceanEM for a UAV, and the
# :class:`open3d.geometry.TriangleMesh` structures can be accessed by importing the data subpackage
import lyceanem.tests.reflectordata as data

body, array, source_coords = data.exampleUAV(10e9)

# %%
# Visualise the Resultant UAV and Array
# ---------------------------------------
# :func:`open3d.visualization.draw_geometries` can be used to visualise the open3d data
# structures :class:`open3d.geometry.PointCloud` and :class:`open3d.geometry.PointCloud`

mesh_frame = o3d.geometry.TriangleMesh.create_coordinate_frame(
    size=0.5, origin=[0, 0, 0]
)
o3d.visualization.draw_geometries([body, array, source_coords, mesh_frame])

# %%
# .. image:: ../_static/UAVArraywithPoints.png


from lyceanem.base_classes import structures

blockers = structures([body, array])

# %%
# Model Farfield Array Patterns
# -------------------------------
# The same function is used to predict the farfield pattern of each element in the array, but the variable 'elements'
# is set as True, instructing the function to return the antenna patterns as 3D arrays arranged with axes element,
# elevation points, and azimuth points. These can then be beamformed using the desired beamforming algorithm. LyceanEM
# currently includes two open loop algorithms for phase weights :func:`lyceanem.electromagnetics.beamforming.EGCWeights`,
# and :func:`lyceanem.electromagnetics.beamforming.WavefrontWeights`
from lyceanem.models.frequency_domain import calculate_farfield

desired_E_axis = np.zeros((1, 3), dtype=np.float32)
desired_E_axis[0, 2] = 1.0

Etheta, Ephi = calculate_farfield(
    source_coords,
    blockers,
    desired_E_axis,
    az_range=np.linspace(-180, 180, az_res),
    el_range=np.linspace(-90, 90, elev_res),
    wavelength=wavelength,
    farfield_distance=20,
    elements=True,
    project_vectors=True,
)


from lyceanem.electromagnetics.beamforming import MaximumDirectivityMap

az_range = np.linspace(-180, 180, az_res)
el_range = np.linspace(-90, 90, elev_res)
directivity_map = MaximumDirectivityMap(
    Etheta, Ephi, source_coords, wavelength, az_range, el_range
)

from lyceanem.electromagnetics.beamforming import PatternPlot

az_mesh, elev_mesh = np.meshgrid(az_range, el_range)

PatternPlot(
    directivity_map[:, :, 2], az_mesh, elev_mesh, logtype="power", plottype="Contour"
)

# %%
# .. image:: ../_static/sphx_glr_05_array_beamforming_001.png

from lyceanem.electromagnetics.beamforming import Steering_Efficiency

setheta, sephi, setot = Steering_Efficiency(
    directivity_map[:, :, 0],
    directivity_map[:, :, 1],
    directivity_map[:, :, 2],
    np.radians(np.diff(el_range)[0]),
    np.radians(np.diff(az_range)[0]),
    4 * np.pi,
)

print("Steering Effciency of {:3.1f}%".format(setot))


print(
    "Maximum Directivity of {:3.1f} dBi".format(
        np.max(10 * np.log10(directivity_map[:, :, 2]))
    )
)



#!/usr/bin/env python
# coding: utf-8
"""
Modelling a Physical Channel in the Time Domain
======================================================

This example uses the frequency domain :func:`lyceanem.models.time_domain.calculate_scattering` function to
predict the time domain response for a given excitation signal and environment included in the model.
This model allows for a very wide range of antennas and antenna arrays to be considered, but for simplicity only horn
antennas will be included in this example. The simplest case would be a single source point and single receive point,
rather than an aperture antenna such as a horn.

"""

import numpy as np
import open3d as o3d
import copy

# %%
# Frequency and Mesh Resolution
# ------------------------------
#
sampling_freq = 60e9
model_time = 1e-7
num_samples = int(model_time * (sampling_freq))

# simulate receiver noise
bandwidth = 8e9
kb = 1.38065e-23
receiver_impedence = 50
thermal_noise_power = 4 * kb * 293.15 * receiver_impedence * bandwidth
noise_power = -80  # dbw
mean_noise = 0

model_freq = 16e9
wavelength = 3e8 / model_freq

# %%
# Setup transmitters and receivers
# -----------------------------------
#
import lyceanem.geometry.targets as TL
import lyceanem.geometry.geometryfunctions as GF

transmit_horn_structure, transmitting_antenna_surface_coords = TL.meshedHorn(
    58e-3, 58e-3, 128e-3, 2e-3, 0.21, wavelength * 0.5
)
receive_horn_structure, receiving_antenna_surface_coords = TL.meshedHorn(
    58e-3, 58e-3, 128e-3, 2e-3, 0.21, wavelength * 0.5
)

# %%
# Position Transmitter
# ----------------------
# rotate the transmitting antenna to the desired orientation, and then translate to final position.
# :func:`lyceanem.geometry.geometryfunctions.open3drotate` allows both the center of rotation to be defined, and
# ensures the right syntax is used for Open3d, as it was changed from 0.9.0 to 0.10.0 and onwards.
#
rotation_vector1 = np.radians(np.asarray([90.0, 0.0, 0.0]))
rotation_vector2 = np.radians(np.asarray([0.0, 0.0, -90.0]))
transmit_horn_structure = GF.open3drotate(
    transmit_horn_structure,
    o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector1),
)
transmit_horn_structure = GF.open3drotate(
    transmit_horn_structure,
    o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector2),
)
transmit_horn_structure.translate(np.asarray([2.695, 0, 0]), relative=True)
transmitting_antenna_surface_coords = GF.open3drotate(
    transmitting_antenna_surface_coords,
    o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector1),
)
transmitting_antenna_surface_coords = GF.open3drotate(
    transmitting_antenna_surface_coords,
    o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector2),
)
transmitting_antenna_surface_coords.translate(np.asarray([2.695, 0, 0]), relative=True)
# %%
# Position Receiver
# ------------------
# rotate the receiving horn to desired orientation and translate to final position.
receive_horn_structure = GF.open3drotate(
    receive_horn_structure,
    o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector1),
)
receive_horn_structure.translate(np.asarray([0, 1.427, 0]), relative=True)
receiving_antenna_surface_coords = GF.open3drotate(
    receiving_antenna_surface_coords,
    o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector1),
)
receiving_antenna_surface_coords.translate(np.asarray([0, 1.427, 0]), relative=True)

# %%
# Create Scattering Plate
# --------------------------
# Create a Scattering plate a source of multipath reflections

reflectorplate, scatter_points = TL.meshedReflector(
    0.3, 0.3, 6e-3, wavelength * 0.5, sides="front"
)
position_vector = np.asarray([29e-3, 0.0, 0])
rotation_vector1 = np.radians(np.asarray([0.0, 90.0, 0.0]))
scatter_points = GF.open3drotate(
    scatter_points,
    o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector1),
)
reflectorplate = GF.open3drotate(
    reflectorplate,
    o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector1),
)
reflectorplate.translate(position_vector, relative=True)
scatter_points.translate(position_vector, relative=True)

# %%
# Specify Reflection Angle
# --------------------------
# Rotate the scattering plate to the optimum angle for reflection from the transmitting to receiving horn

plate_orientation_angle = 45.0

rotation_vector = np.radians(np.asarray([0.0, 0.0, plate_orientation_angle]))
scatter_points = GF.open3drotate(
    scatter_points,
    o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector),
)
reflectorplate = GF.open3drotate(
    reflectorplate,
    o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector),
)

from lyceanem.base_classes import structures

blockers = structures([reflectorplate, receive_horn_structure, transmit_horn_structure])

# %%
# Visualise the Scene Geometry
# ------------------------------
# Use open3d function :func:`open3d.visualization.draw_geometries` to visualise the scene and ensure that all the
# relavent sources and scatter points are correct. Point normal vectors can be displayed by pressing 'n' while the
# window is open.
mesh_frame = o3d.geometry.TriangleMesh.create_coordinate_frame(
    size=0.5, origin=[0, 0, 0]
)
o3d.visualization.draw_geometries(
    [
        transmitting_antenna_surface_coords,
        receiving_antenna_surface_coords,
        scatter_points,
        reflectorplate,
        mesh_frame,
        receive_horn_structure,
        transmit_horn_structure,
    ]
)
# %%
# .. image:: ../_static/03_frequency_domain_channel_model_picture_01.png

# %%
# Specify desired Transmit Polarisation
# --------------------------------------
# The transmit polarisation has a significant effect on the channel characteristics. In this example the transmit horn will be vertically polarised, (e-vector aligned with the z direction)

desired_E_axis = np.zeros((1, 3), dtype=np.float32)
desired_E_axis[0, 1] = 1.0

# %%
# Time Domain Scattering
# ----------------------------
#
import scipy.signal as sig
import lyceanem.models.time_domain as TD
from lyceanem.base_classes import structures


angle_values = np.linspace(0, 90, 91)
angle_increment = np.diff(angle_values)[0]
responsex = np.zeros((len(angle_values)), dtype="complex")
responsey = np.zeros((len(angle_values)), dtype="complex")
responsez = np.zeros((len(angle_values)), dtype="complex")

plate_orientation_angle = -45.0

rotation_vector = np.radians(
    np.asarray([0.0, 0.0, plate_orientation_angle + angle_increment])
)
scatter_points = GF.open3drotate(
    scatter_points,
    o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector),
)
reflectorplate = GF.open3drotate(
    reflectorplate,
    o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector),
)

from tqdm import tqdm

wake_times = np.zeros((len(angle_values)))
Ex = np.zeros((len(angle_values), num_samples))
Ey = np.zeros((len(angle_values), num_samples))
Ez = np.zeros((len(angle_values), num_samples))

for angle_inc in tqdm(range(len(angle_values))):
    rotation_vector = np.radians(np.asarray([0.0, 0.0, angle_increment]))
    scatter_points = GF.open3drotate(
        scatter_points,
        o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector),
    )
    reflectorplate = GF.open3drotate(
        reflectorplate,
        o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector),
    )
    blockers = structures(
        [reflectorplate, transmit_horn_structure, receive_horn_structure]
    )
    pulse_time = 5e-9
    output_power = 0.01  # dBwatts
    powerdbm = 10 * np.log10(output_power) + 30
    v_transmit = ((10 ** (powerdbm / 20)) * receiver_impedence) ** 0.5
    output_amplitude_rms = v_transmit / (1 / np.sqrt(2))
    output_amplitude_peak = v_transmit

    desired_E_axis = np.zeros((3), dtype=np.float32)
    desired_E_axis[2] = 1.0
    noise_volts_peak = (10 ** (noise_power / 10) * receiver_impedence) * 0.5

    excitation_signal = output_amplitude_rms * sig.chirp(
        np.linspace(0, pulse_time, int(pulse_time * sampling_freq)),
        model_freq - bandwidth,
        pulse_time,
        model_freq,
        method="linear",
        phi=0,
        vertex_zero=True,
    ) + np.random.normal(mean_noise, noise_volts_peak, int(pulse_time * sampling_freq))
    (
        Ex[angle_inc, :],
        Ey[angle_inc, :],
        Ez[angle_inc, :],
        wake_times[angle_inc],
    ) = TD.calculate_scattering(
        transmitting_antenna_surface_coords,
        receiving_antenna_surface_coords,
        excitation_signal,
        blockers,
        desired_E_axis,
        scatter_points=scatter_points,
        wavelength=wavelength,
        scattering=1,
        elements=False,
        sampling_freq=sampling_freq,
        num_samples=num_samples,
    )

    noise_volts = np.random.normal(mean_noise, noise_volts_peak, num_samples)
    Ex[angle_inc, :] = Ex[angle_inc, :] + noise_volts
    Ey[angle_inc, :] = Ey[angle_inc, :] + noise_volts
    Ez[angle_inc, :] = Ez[angle_inc, :] + noise_volts


# %%
# Plot Normalised Response
# ----------------------------
# Using matplotlib, plot the results


import matplotlib.pyplot as plt

time_index = np.linspace(0, model_time * 1e9, num_samples)
time, anglegrid = np.meshgrid(time_index[:1801], angle_values - 45)
norm_max = np.nanmax(
    np.array(
        [
            np.nanmax(10 * np.log10((Ex ** 2) / receiver_impedence)),
            np.nanmax(10 * np.log10((Ey ** 2) / receiver_impedence)),
            np.nanmax(10 * np.log10((Ez ** 2) / receiver_impedence)),
        ]
    )
)

fig2, ax2 = plt.subplots(constrained_layout=True)
origin = "lower"
# Now make a contour plot with the levels specified,
# and with the colormap generated automatically from a list
# of colors.

levels = np.linspace(-80, 0, 41)

CS = ax2.contourf(
    anglegrid,
    time,
    10 * np.log10((Ez[:, :1801] ** 2) / receiver_impedence) - norm_max,
    levels,
    origin=origin,
    extend="both",
)
cbar = fig2.colorbar(CS)
cbar.ax.set_ylabel("Received Power (dBm)")

ax2.set_ylim(0, 30)
ax2.set_xlim(-45, 45)

ax2.set_xticks(np.linspace(-45, 45, 7))
ax2.set_yticks(np.linspace(0, 30, 16))

ax2.set_xlabel("Rotation Angle (degrees)")
ax2.set_ylabel("Time of Flight (ns)")
ax2.set_title("Received Power vs Time for rotating Plate (24GHz)")

from scipy.fft import fft, fftfreq
import scipy

xf = fftfreq(len(time_index), 1 / sampling_freq)[: len(time_index) // 2]
input_signal = excitation_signal * (output_amplitude_peak)
inputfft = fft(input_signal)
input_freq = fftfreq(120, 1 / sampling_freq)[:60]
freqfuncabs = scipy.interpolate.interp1d(input_freq, np.abs(inputfft[:60]))
freqfuncangle = scipy.interpolate.interp1d(input_freq, np.angle(inputfft[:60]))
newinput = freqfuncabs(xf[1600]) * np.exp(freqfuncangle(xf[1600]))
Exf = fft(Ex)
Eyf = fft(Ey)
Ezf = fft(Ez)

# %%
# .. image:: ../_static/sphx_glr_04_time_domain_channel_modelling_001.png

# %%
# Frequency Specific Results
# -------------------------------
# The time of flight plot is useful to displaying the output of the model, giving a understanding about what is
# physically happening in the channel, but to get an idea of the behaviour in the frequency domain we need to use a
# fourier transform to move from time and voltages to frequency.

s21x = 20 * np.log10(np.abs(Exf[:, 1600] / newinput))
s21y = 20 * np.log10(np.abs(Eyf[:, 1600] / newinput))
s21z = 20 * np.log10(np.abs(Ezf[:, 1600] / newinput))
tdangles = np.linspace(-45, 45, 91)
fig, ax = plt.subplots()
ax.plot(tdangles, s21x - np.max(s21z), label="Ex")
ax.plot(tdangles, s21y - np.max(s21z), label="Ey")
ax.plot(tdangles, s21z - np.max(s21z), label="Ez")
plt.xlabel("$\\theta_{N}$ (degrees)")
plt.ylabel("Normalised Level (dB)")
ax.set_ylim(-60.0, 0)
ax.set_xlim(np.min(angle_values) - 45, np.max(angle_values) - 45)
ax.set_xticks(np.linspace(np.min(angle_values) - 45, np.max(angle_values) - 45, 19))
ax.set_yticks(np.linspace(-60, 0.0, 21))
legend = ax.legend(loc="upper right", shadow=True)
plt.grid()
plt.title("$S_{21}$ at 16GHz")
plt.show()

# %%
# .. image:: ../_static/sphx_glr_04_time_domain_channel_modelling_002.png
